import {DOCUMENT} from "@angular/common";
import {
  Component,
  HostListener,
  Inject,
  Renderer2,
} from "@angular/core";
import {
  Router,
  RouterStateSnapshot,
} from "@angular/router";
import {SharedAbstractPresentational} from "@app/shared/components/presentationals/shared-abstract/shared-abstract.presentational";
import {ApplicationRoleEnum} from "@app/shared/enums/application-role.enum";
import {LanguagesEnum} from "@app/shared/enums/languages.enum";
import {
  RoutesEnum,
  urlSeparator,
} from "@app/shared/enums/routes.enum";
import {ThemeEnum} from "@app/shared/enums/theme.enum";
import {LocalStateModel} from "@app/shared/models/local-state.model";
import {AppAction} from "@app/stores/app.action";
import {TranslateService} from "@ngx-translate/core";
import {Navigate} from "@ngxs/router-plugin";
import {
  Select,
  Store,
} from "@ngxs/store";
import {UserExtended} from "@shared/models/business/user-extended.model";
import {AppStatusService} from "@shared/services/app-status.service";
import {Observable} from "rxjs";
import {tap} from "rxjs/internal/operators/tap";
import {distinctUntilChanged} from "rxjs/operators";
import {
  isFalse,
  isFalsy,
  isTrue,
  isUndefined,
  NotificationService,
  OAuth2Service,
  TRANSLATE,
} from "solidify-frontend";
import {environment} from "../environments/environment";

@Component({
  selector: "dlcm-root",
  templateUrl: "./app.component.html",
  styleUrls: ["./app.component.scss"],
})
export class AppComponent extends SharedAbstractPresentational {

  private static readonly _themeAttributeName: string = "theme";
  currentModule: string;
  logo: string;

  @Select((state: LocalStateModel) => state.router.state) urlStateObs: Observable<RouterStateSnapshot>;
  @Select((state: LocalStateModel) => state.application.appLanguage) appLanguageObs: Observable<LanguagesEnum>;
  @Select((state: LocalStateModel) => state.application.isLoggedIn) isLoggedObs: Observable<boolean>;
  @Select((state: LocalStateModel) => state.application.isApplicationInitialized) isApplicationInitializedObs: Observable<boolean>;
  @Select((state: LocalStateModel) => state.application.theme) themeObs: Observable<ThemeEnum>;
  @Select((state: LocalStateModel) => state.application.userRoles) userRolesObs: Observable<ApplicationRoleEnum[]>;
  @Select((state: LocalStateModel) => state.application.application_user.current) currentUserObs: Observable<UserExtended>;
  ignoreGrid: boolean = false;

  @HostListener("window:beforeunload", ["$event"])
  leaveWithoutSaveNotification($event: Event): void {
    const preventExit = this.store.selectSnapshot((state: LocalStateModel) => state.application.preventExit);
    const ignorePreventExit = this.store.selectSnapshot((state: LocalStateModel) => state.application.ignorePreventLeavePopup);
    if (isFalse(ignorePreventExit) && isTrue(preventExit)) {
      $event.returnValue = true;
    }
  }

  constructor(private store: Store,
              private oauthService: OAuth2Service,
              private router: Router,
              private translate: TranslateService,
              private _renderer: Renderer2,
              @Inject(DOCUMENT) private readonly _document: Document,
              private _notificationService: NotificationService,
              public appStatusService: AppStatusService) {
    super();
    this.store.dispatch(new AppAction.InitApplication());
    this.setTitle();
    this.listenCurrentModule();
    this.listenThemeChange();
    this.observeOfflineOnlineMode();
  }

  private observeOfflineOnlineMode(): void {
    let previousStateIsOnline: boolean | undefined = undefined;
    this.subscribe(this.appStatusService.online.pipe(
      distinctUntilChanged(),
      tap((isOnline: boolean) => {
        if ((isUndefined(previousStateIsOnline) || isTrue(previousStateIsOnline)) && isFalsy(isOnline)) {
          this._notificationService.showWarning(TRANSLATE("app.status.offline"), true);
          previousStateIsOnline = isOnline;
          return;
        }

        if (isFalse(previousStateIsOnline) && isTrue(isOnline)) {
          this._notificationService.showInformation(TRANSLATE("app.status.backToOnline"), true);
          previousStateIsOnline = isOnline;
          return;
        }
        return;
      }),
    ));
  }

  useLanguage(language: LanguagesEnum): void {
    this.store.dispatch(new AppAction.ChangeAppLanguage(language));
  }

  navigate(route: string): void {
    this.store.dispatch(new Navigate([route]));
  }

  private setTitle(): void {
    this._document.title = environment.appTitle;
  }

  private setTheme(): void {
    this._renderer.setAttribute(this._document.body, AppComponent._themeAttributeName, environment.theme);
  }

  private setFavicon(): void {
    const link = this._document.querySelector("link[rel*='icon']") || this._document.createElement("link");
    this._renderer.setAttribute(link, "type", "image/x-icon");
    this._renderer.setAttribute(link, "rel", "shortcut icon");
    this._renderer.setAttribute(link, "href", `./assets/themes/${environment.theme}/favicon.ico`);
  }

  private setImageToolbar(): void {
    this.logo = `assets/themes/${environment.theme}/toolbar-header-image.svg`;
  }

  private listenCurrentModule(): void {
    this.subscribe(this.urlStateObs
      .pipe(
        tap(urlState => {
          this.ignoreGrid = false;
          if (urlState) {
            const url = urlState.url;
            if (url.startsWith(urlSeparator + RoutesEnum.homePage)) {
              this.currentModule = RoutesEnum.homePage;

              if (url.startsWith(urlSeparator + RoutesEnum.homePageBrowsing)) {
                this.ignoreGrid = true;
              }
              return;
            }

            if (url.startsWith(urlSeparator + RoutesEnum.deposit)) {
              this.currentModule = RoutesEnum.deposit;
              return;
            }

            if (url.startsWith(urlSeparator + RoutesEnum.organizationalUnit)) {
              this.currentModule = RoutesEnum.organizationalUnit;
              return;
            }

            if (url.startsWith(urlSeparator + RoutesEnum.admin)) {
              this.currentModule = RoutesEnum.admin;
              return;
            }

            this.currentModule = null;
          }
        }),
      ),
    );
  }

  logout(): void {
    this.store.dispatch(new AppAction.Logout());
  }

  getTitle(): string {
    return this._document.title;
  }

  themeChange(theme: ThemeEnum): void {
    this.store.dispatch(new AppAction.ChangeAppTheme(theme));
  }

  navigateToHome(): void {
    this.navigate(RoutesEnum.homePage);
  }

  private listenThemeChange(): void {
    this.subscribe(
      this.themeObs.pipe(
        tap((theme: ThemeEnum) => {
          this.setTheme();
          this.setFavicon();
          this.setImageToolbar();
        }),
      ),
    );
  }
}
