import {
  HTTP_INTERCEPTORS,
  HttpClient,
  HttpClientModule,
} from "@angular/common/http";
import {
  APP_INITIALIZER,
  ErrorHandler,
  NgModule,
} from "@angular/core";
import {
  DateAdapter,
  MAT_DATE_FORMATS,
  MAT_DATE_LOCALE,
  MatPaginatorIntl,
  MatSnackBar,
} from "@angular/material";
import {
  MAT_MOMENT_DATE_ADAPTER_OPTIONS,
  MAT_MOMENT_DATE_FORMATS,
  MomentDateAdapter,
} from "@angular/material-moment-adapter";
import {BrowserModule} from "@angular/platform-browser";
import {BrowserAnimationsModule} from "@angular/platform-browser/animations";
import {ServiceWorkerModule} from "@angular/service-worker";
import {AppRoutingModule} from "@app/app-routing.module";
import {TokenDialog} from "@app/components/dialogs/token/token.dialog";
import {UserDialog} from "@app/components/dialogs/user/user.dialog";
import {AvatarPresentational} from "@app/components/presentationals/avatar/avatar.presentational";
import {MainToolbarDesktopVerticalPresentational} from "@app/components/presentationals/main-toolbar/main-toolbar-desktop-vertical/main-toolbar-desktop-vertical.presentational";
import {MainToolbarMobilePresentational} from "@app/components/presentationals/main-toolbar/main-toolbar-mobile/main-toolbar-mobile.presentational";
import {ThemeSelectorPresentational} from "@app/components/presentationals/theme-selector/theme-selector.presentational";
import {UserForm} from "@app/components/presentationals/user-form/user-form";
import {InMemoryStorage} from "@app/in-memory.storage";
import {DlcmErrorsHandlerService} from "@app/shared/services/dlcm-errors-handler.service";
import {SharedModule} from "@app/shared/shared.module";
import {AppState} from "@app/stores/app.state";
import {AppAuthorizedOrganizationalUnitState} from "@app/stores/authorized-organizational-unit/app-authorized-organizational-unit.state";
import {AppOrganizationalUnitPersonRoleState} from "@app/stores/organizational-unit-person-role/app-organizational-unit-person-role.state";
import {AppPersonState} from "@app/stores/person/app-person.state";
import {AppUserState} from "@app/stores/user/app-user.state";
import {
  TranslateLoader,
  TranslateModule,
} from "@ngx-translate/core";
import {TranslateHttpLoader} from "@ngx-translate/http-loader";
import {NgxsReduxDevtoolsPluginModule} from "@ngxs/devtools-plugin";
import {NgxsLoggerPluginModule} from "@ngxs/logger-plugin";
import {NgxsRouterPluginModule} from "@ngxs/router-plugin";
import {NgxsModule} from "@ngxs/store";
import {CustomMatPaginatorIntlService} from "@shared/services/custom-mat-paginator-intl.service";
import xml from "highlight.js/lib/languages/xml";
import {HighlightModule} from "ngx-highlightjs";
import {
  AppConfigService,
  ENVIRONMENT,
  OAuth2Interceptor,
  OAuthStorage,
  SNACK_BAR,
} from "solidify-frontend";
import {environment} from "../environments/environment";
import {AppComponent} from "./app.component";
import {FooterPresentational} from "./components/presentationals/footer/footer.presentational";
import {LanguageSelectorPresentational} from "./components/presentationals/language-selector/language-selector.presentational";
import {MainToolbarDesktopHorizontalPresentational} from "./components/presentationals/main-toolbar/main-toolbar-desktop/main-toolbar-desktop-horizontal-presentational.component";
import {PageNotFoundPresentational} from "./components/presentationals/page-not-found/page-not-found.presentational";

const presentationals = [
  AppComponent,
  FooterPresentational,
  LanguageSelectorPresentational,
  MainToolbarDesktopHorizontalPresentational,
  MainToolbarDesktopVerticalPresentational,
  MainToolbarMobilePresentational,
  PageNotFoundPresentational,
  ThemeSelectorPresentational,
  AvatarPresentational,
  UserForm,
];

const dialogs = [
  TokenDialog,
  UserDialog,
];

const appInitializerFn = (appConfig: AppConfigService) => () => appConfig.mergeConfig(environment).toPromise();

export const WINDOW = "windowObject";

export function createDefaultStorage(): Storage | null {
  return environment.tokenInMemoryStorage ? new InMemoryStorage() : (typeof sessionStorage !== "undefined" ? sessionStorage : null);
}

export function hljsLanguages(): any {
  return [
    {name: "xml", func: xml},
  ];
}

@NgModule({
  declarations: [
    ...presentationals,
    ...dialogs,
  ],
  imports: [
    // angular
    BrowserModule,
    BrowserAnimationsModule,
    HttpClientModule,

    // NgXs
    NgxsModule.forRoot([
      AppState,
      AppUserState,
      AppPersonState,
      AppAuthorizedOrganizationalUnitState,
      AppOrganizationalUnitPersonRoleState,
    ], {
      developmentMode: !environment.production, // Allow to enable freeze store in dev env
    }),
    NgxsLoggerPluginModule.forRoot({
      logger: console,
      collapsed: false,
      disabled: environment.production,
    }),
    NgxsReduxDevtoolsPluginModule.forRoot({
      disabled: environment.production,
    }),
    NgxsRouterPluginModule.forRoot(),

    // NgTranslate
    TranslateModule.forRoot({
      loader: {
        provide: TranslateLoader,
        useFactory: HttpLoaderFactory,
        deps: [HttpClient],
      },
    }),

    // Shared
    SharedModule,

    // App
    AppRoutingModule,
    HighlightModule.forRoot({
      languages: hljsLanguages,
    }),
    ServiceWorkerModule.register("./ngsw-worker.js", {enabled: environment.production, registrationStrategy: "registerImmediately"}),
  ],
  providers: [
    {
      provide: ENVIRONMENT,
      useValue: environment,
    },
    {
      provide: APP_INITIALIZER,
      useFactory: appInitializerFn,
      multi: true,
      deps: [AppConfigService],
    },
    {
      provide: HTTP_INTERCEPTORS,
      useExisting: OAuth2Interceptor,
      multi: true,
    },
    {
      provide: OAuthStorage,
      useFactory: createDefaultStorage,
    },
    {
      provide: MatPaginatorIntl,
      useClass: CustomMatPaginatorIntlService,
    },
    {
      provide: ErrorHandler,
      useClass: DlcmErrorsHandlerService,
    },
    {
      provide: SNACK_BAR,
      useClass: MatSnackBar,
    },
    {
      provide: WINDOW,
      useValue: window,
    },
    {
      provide: MAT_MOMENT_DATE_ADAPTER_OPTIONS,
      useValue: {
        useUtc: true,
      },
    },
    {provide: MAT_DATE_LOCALE, useValue: navigator.language},
    {provide: DateAdapter, useClass: MomentDateAdapter, deps: [MAT_DATE_LOCALE]},
    {provide: MAT_DATE_FORMATS, useValue: MAT_MOMENT_DATE_FORMATS},

  ],
  entryComponents: [
    ...dialogs,
  ],
  bootstrap: [AppComponent],
})
export class AppModule {
}

// required for AOT compilation
export function HttpLoaderFactory(http: HttpClient): TranslateHttpLoader {
  return new TranslateHttpLoader(http, "./assets/i18n/");
}
