import {PersonExtended} from "@admin/models/person-extended.model";
import {PersonRole} from "@admin/models/person-role.model";
import {
  ChangeDetectionStrategy,
  ChangeDetectorRef,
  Component,
  Input,
} from "@angular/core";
import {
  FormBuilder,
  Validators,
} from "@angular/forms";
import {
  FundingAgency,
  Institution,
  PreservationPolicy,
  Role,
  SubmissionPolicy,
} from "@app/generated-api";
import {OrganizationalUnitExtended} from "@deposit/models/organizational-unit-extended.model";
import {SharedAbstractFormPresentational} from "@shared/components/presentationals/shared-abstract-form/shared-abstract-form.presentational";
import {PersonOrgUnitRoleMode} from "@shared/components/presentationals/shared-person-orgunit-role/shared-person-orgunit-role.presentational";
import {LocalModelAttributeEnum} from "@shared/enums/model-attribute.enum";
import {BaseFormDefinition} from "@shared/models/base-form-definition.model";
import {BreakpointService} from "@shared/services/breakpoint.service";
import _ from "lodash";
import {
  DateUtil,
  PropertyName,
  SolidifyValidator,
} from "solidify-frontend";

@Component({
  selector: "dlcm-admin-orgunit-form",
  templateUrl: "./admin-orgunit-form.presentational.html",
  styleUrls: ["../../../../../../shared/components/presentationals/shared-abstract-form/shared-abstract-form.presentational.scss"],
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class AdminOrgunitFormPresentational extends SharedAbstractFormPresentational<OrganizationalUnitExtended> {
  formDefinition: FormComponentFormDefinition = new FormComponentFormDefinition();

  @Input()
  selectedSubmissionPolicies: SubmissionPolicy[];

  @Input()
  selectedPreservationPolicies: PreservationPolicy[];

  @Input()
  selectedFundingAgencies: FundingAgency[];

  @Input()
  selectedPersonRole: PersonRole[];

  @Input()
  selectedInstitutions: Institution[];

  @Input()
  listSubmissionPolicies: SubmissionPolicy[];

  @Input()
  listPreservationPolicies: PreservationPolicy[];

  @Input()
  listFundingAgencies: FundingAgency[];

  @Input()
  listInstitutions: Institution[];

  @Input()
  listPerson: PersonExtended[];

  @Input()
  listRole: Role[];

  get personOrgUnitRoleMode(): typeof PersonOrgUnitRoleMode {
    return PersonOrgUnitRoleMode;
  }

  constructor(protected readonly _changeDetectorRef: ChangeDetectorRef,
              private readonly _fb: FormBuilder,
              public readonly breakpointService: BreakpointService) {
    super(_changeDetectorRef);
  }

  protected initNewForm(): void {
    this.form = this._fb.group({
      [this.formDefinition.name]: ["", [Validators.required, SolidifyValidator]],
      [this.formDefinition.description]: ["", [SolidifyValidator]],
      [this.formDefinition.openingDate]: [""],
      [this.formDefinition.closingDate]: [""],
      [this.formDefinition.submissionPolicies]: [[], [Validators.required, SolidifyValidator]],
      [this.formDefinition.preservationPolicies]: [[], [Validators.required, SolidifyValidator]],
      [this.formDefinition.defaultSubmissionPolicy]: [undefined, [Validators.required, SolidifyValidator]],
      [this.formDefinition.defaultPreservationPolicy]: [undefined, [Validators.required, SolidifyValidator]],
      [this.formDefinition.url]: ["", [SolidifyValidator]],
      [this.formDefinition.personRole]: ["", [SolidifyValidator]],
      [this.formDefinition.fundingAgencies]: ["", [SolidifyValidator]],
      [this.formDefinition.institutions]: ["", [SolidifyValidator]]
    });
  }

  protected bindFormTo(organizationalUnit: OrganizationalUnitExtended): void {
    this.form = this._fb.group({
      [this.formDefinition.name]: [organizationalUnit.name, [Validators.required, SolidifyValidator]],
      [this.formDefinition.description]: [organizationalUnit.description, [SolidifyValidator]],
      [this.formDefinition.openingDate]: [organizationalUnit.openingDate],
      [this.formDefinition.closingDate]: [organizationalUnit.closingDate],
      [this.formDefinition.submissionPolicies]: [_.map(this.selectedSubmissionPolicies, LocalModelAttributeEnum.resId), [Validators.required, SolidifyValidator]],
      [this.formDefinition.preservationPolicies]: [_.map(this.selectedPreservationPolicies, LocalModelAttributeEnum.resId), [Validators.required, SolidifyValidator]],
      [this.formDefinition.defaultSubmissionPolicy]: [(organizationalUnit.defaultSubmissionPolicy ? organizationalUnit.defaultSubmissionPolicy.resId : undefined), [Validators.required, SolidifyValidator]],
      [this.formDefinition.defaultPreservationPolicy]: [(organizationalUnit.defaultPreservationPolicy ? organizationalUnit.defaultPreservationPolicy.resId : undefined), [Validators.required, SolidifyValidator]],
      [this.formDefinition.url]: [organizationalUnit.url, [SolidifyValidator]],
      [this.formDefinition.personRole]: ["", [SolidifyValidator]],
      [this.formDefinition.fundingAgencies]: [_.map(this.selectedFundingAgencies, LocalModelAttributeEnum.resId), [SolidifyValidator]],
      [this.formDefinition.institutions]: [_.map(this.selectedInstitutions, LocalModelAttributeEnum.resId), [SolidifyValidator]],
    });
  }

  protected treatmentBeforeSubmit(organizationalUnit: OrganizationalUnitExtended | any): OrganizationalUnitExtended {
    organizationalUnit.openingDate = DateUtil.convertToLocalDateDateSimple(organizationalUnit.openingDate);
    organizationalUnit.closingDate = DateUtil.convertToLocalDateDateSimple(organizationalUnit.closingDate);
    organizationalUnit.defaultSubmissionPolicy = {resId: organizationalUnit[this.formDefinition.defaultSubmissionPolicy]};
    organizationalUnit.defaultPreservationPolicy = {resId: organizationalUnit[this.formDefinition.defaultPreservationPolicy]};
    organizationalUnit.preservationPolicies = [];
    this.form.get(this.formDefinition.preservationPolicies).value.forEach(resId => {
      organizationalUnit.preservationPolicies.push({resId: resId});
    });
    organizationalUnit.submissionPolicies = [];
    this.form.get(this.formDefinition.submissionPolicies).value.forEach(resId => {
      organizationalUnit.submissionPolicies.push({resId: resId});
    });
    organizationalUnit.fundingAgencies = [];
    if (this.form.get(this.formDefinition.fundingAgencies).value !== "") {
      this.form.get(this.formDefinition.fundingAgencies).value.forEach(resId => {
        organizationalUnit.fundingAgencies.push({resId: resId});
      });
    }
    organizationalUnit.institutions = [];
    if (this.form.get(this.formDefinition.institutions).value !== "") {
      this.form.get(this.formDefinition.institutions).value.forEach(resId => {
        organizationalUnit.institutions.push({resId: resId});
      });
    }
    return organizationalUnit;
  }
}

class FormComponentFormDefinition extends BaseFormDefinition {
  @PropertyName() name: string;
  @PropertyName() description: string;
  @PropertyName() closingDate: string;
  @PropertyName() openingDate: string;
  @PropertyName() submissionPolicies: string;
  @PropertyName() preservationPolicies: string;
  @PropertyName() defaultSubmissionPolicy: string;
  @PropertyName() defaultPreservationPolicy: string;
  @PropertyName() url: string;
  @PropertyName() personRole: string;
  @PropertyName() fundingAgencies: string;
  @PropertyName() institutions: string;
}
