import {
  ChangeDetectionStrategy,
  Component,
  OnInit,
} from "@angular/core";
import {
  FormBuilder,
  FormGroup,
  Validators,
} from "@angular/forms";
import {MatDialogRef} from "@angular/material";
import {FileUploadWrapper} from "@app/features/deposit/models/file-upload-wrapper.model";
import {SharedAbstractContainer} from "@app/shared/components/containers/shared-abstract/shared-abstract.container";
import {DataCategoryEnum} from "@app/shared/enums/data-category.enum";
import {DataTypeEnum} from "@app/shared/enums/data-type.enum";
import {DataCategoryHelper} from "@app/shared/helpers/data-category.helper";
import {BaseFormDefinition} from "@app/shared/models/base-form-definition.model";
import {PropertyName} from "solidify-frontend";

@Component({
  selector: "dlcm-deposit-file-upload-dialog",
  templateUrl: "./deposit-file-upload.dialog.html",
  styleUrls: ["./deposit-file-upload.dialog.scss"],
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class DepositFileUploadDialog extends SharedAbstractContainer implements OnInit {
  form: FormGroup;
  formDefinition: FormComponentFormDefinition = new FormComponentFormDefinition();
  filesToUpload: File[] = [];

  get dataCategoryHelper(): typeof DataCategoryHelper {
    return DataCategoryHelper;
  }

  readonly files: string = "files";

  constructor(protected _dialogRef: MatDialogRef<DepositFileUploadDialog>,
              protected _fb: FormBuilder) {
    super();
  }

  ngOnInit(): void {
    super.ngOnInit();
    this.form = this._fb.group({
      [this.formDefinition.subDirectory]: [""],
      [this.formDefinition.dataCategory]: [DataCategoryEnum.Primary, [Validators.required]],
      [this.formDefinition.dataType]: [DataTypeEnum.Reference, [Validators.required]],
    });
  }

  close(): void {
    this._dialogRef.close();
  }

  onFileChange(event: Event & any): void {
    if (event.target.files.length > 0) {
      const files: FileList = event.target.files;
      this.addNewFileToExistingList(this.convertFileListToArray(files));
    }
  }

  convertFileListToArray(files: FileList): File[] {
    const arrayFile: File[] = [];
    for (let i = 0; i < files.length; i++) {
      arrayFile.push(files.item(i));
    }
    return arrayFile;
  }

  addNewFileToExistingList(files: File[]): void {
    files = files.filter(fileToFilter => {
      const isIncluded = this.getIndexOfFile(fileToFilter) !== -1;
      return !isIncluded;
    });
    this.filesToUpload.push(...files);
  }

  private getIndexOfFile(file: File): number {
    return this.filesToUpload.findIndex(fileAlreadyToUpload => fileAlreadyToUpload.name === file.name && fileAlreadyToUpload.type === file.type && fileAlreadyToUpload.size === file.size && fileAlreadyToUpload.lastModified === file.lastModified);
  }

  getDate(lastModified: number): Date {
    return new Date(lastModified);
  }

  onSubmit(): void {
    const base = this.form.value as FileUploadWrapper;
    const listFilesUploadWrapper: FileUploadWrapper[] = [];
    this.filesToUpload.forEach(f => {
      const fileUploadWrapper = {} as FileUploadWrapper;
      Object.assign(fileUploadWrapper, base);
      fileUploadWrapper.file = f;
      listFilesUploadWrapper.push(fileUploadWrapper);
    });
    this._dialogRef.close(listFilesUploadWrapper);
  }

  getDataCategory(): DataCategoryEnum {
    return this.form.get(this.formDefinition.dataCategory).value;
  }

  delete(file: File): void {
    this.filesToUpload.splice(this.getIndexOfFile(file), 1);
  }
}

class FormComponentFormDefinition extends BaseFormDefinition {
  @PropertyName() subDirectory: string;
  @PropertyName() dataCategory: string;
  @PropertyName() dataType: string;
}

