import {NestedTreeControl} from "@angular/cdk/tree";
import {
  ChangeDetectionStrategy,
  Component,
  Input,
  Output,
} from "@angular/core";
import {
  MatDialog,
  MatTreeNestedDataSource,
} from "@angular/material";
import {DepositDataFile} from "@app/features/deposit/models/deposit-data-file.model";
import {SharedAbstractPresentational} from "@app/shared/components/presentationals/shared-abstract/shared-abstract.presentational";
import {
  BehaviorSubject,
  Observable,
} from "rxjs";
import {
  isString,
  ObservableUtil,
  StringUtil,
} from "solidify-frontend";

interface FileNode {
  name: string;
  file?: DepositDataFile;
  children?: FileNode[];
}

@Component({
  selector: "dlcm-deposit-file-tree",
  templateUrl: "./deposit-file-tree.presentational.html",
  styleUrls: ["./deposit-file-tree.presentational.scss"],
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class DepositFileTreePresentational extends SharedAbstractPresentational {
  private _listDepositDataFile: DepositDataFile[];

  @Input()
  readonly: boolean;

  @Input()
  isLoading: boolean;

  @Input()
  isLoadingHistory: boolean;

  @Input()
  set listDepositDataFile(val: DepositDataFile[]) {
    this._listDepositDataFile = val;
    this.adaptListDepositDataFileToTree();
    this.applyTreeToDataSource();
  }

  get listDepositDataFile(): DepositDataFile[] {
    return this._listDepositDataFile;
  }

  private readonly _resumeBS: BehaviorSubject<DepositDataFile | undefined> = new BehaviorSubject<DepositDataFile | undefined>(undefined);
  @Output("resumeChange")
  readonly resumeObs: Observable<DepositDataFile | undefined> = ObservableUtil.asObservable(this._resumeBS);

  private readonly _deleteBS: BehaviorSubject<DepositDataFile | undefined> = new BehaviorSubject<DepositDataFile | undefined>(undefined);
  @Output("deleteChange")
  readonly deleteObs: Observable<DepositDataFile | undefined> = ObservableUtil.asObservable(this._deleteBS);

  private readonly _refreshBS: BehaviorSubject<void | undefined> = new BehaviorSubject<void | undefined>(undefined);
  @Output("refreshChange")
  readonly refreshObs: Observable<void | undefined> = ObservableUtil.asObservable(this._refreshBS);

  private readonly _downloadBS: BehaviorSubject<DepositDataFile | undefined> = new BehaviorSubject<DepositDataFile | undefined>(undefined);
  @Output("downloadChange")
  readonly downloadObs: Observable<DepositDataFile | undefined> = ObservableUtil.asObservable(this._downloadBS);

  private readonly _showHistoryBS: BehaviorSubject<DepositDataFile | undefined> = new BehaviorSubject<DepositDataFile | undefined>(undefined);
  @Output("showHistoryChange")
  readonly showHistoryObs: Observable<DepositDataFile | undefined> = ObservableUtil.asObservable(this._showHistoryBS);

  fileTree: FileNode[];

  treeControl: NestedTreeControl<FileNode> = new NestedTreeControl<FileNode>(node => node.children);
  dataSource: MatTreeNestedDataSource<FileNode> = new MatTreeNestedDataSource<FileNode>();

  private readonly _SEPARATOR: string = "/";
  private readonly _ROOT: string = this._SEPARATOR;

  constructor(public dialog: MatDialog) {
    super();
  }

  adaptListDepositDataFileToTree(): void {
    this.extractDataFile();
    const copy = [...this.listDepositDataFile];
    copy.sort((a, b) => {
      if (a.relativeLocation < b.relativeLocation) {
        return -1;
      }
      if (a.relativeLocation > b.relativeLocation) {
        return 1;
      }
      return 0;
    });
    copy.forEach(dataFile => {
      this.treatmentDataFile(dataFile);
    });
  }

  private extractDataFile(): void {
    this.fileTree = [];
  }

  private treatmentDataFile(dataFile: DepositDataFile): void {
    const isIntoRootFolder = dataFile.relativeLocation === this._ROOT;
    if (isIntoRootFolder) {
      this.addFileToRootDataFile(dataFile);
    } else {
      const relativeLocation = dataFile.relativeLocation;
      const pathParts = isString(relativeLocation) ? relativeLocation.split(this._SEPARATOR)
        .filter(part => part !== StringUtil.stringEmpty) : [];
      this.recursiveAddFolder(this.fileTree, dataFile, pathParts);
    }
  }

  private recursiveAddFolder(levelParent: FileNode[], dataFile: DepositDataFile, pathParts: string[]): void {
    if (pathParts.length === 0) {
      return;
    }
    const firstFolderName = pathParts.splice(0, 1)[0];
    let parentFileNode = levelParent.find(f => f.name === firstFolderName);

    parentFileNode = this.createLevelIfNotExist(parentFileNode, firstFolderName, levelParent);
    this.initChildrenAttributIfUndefined(parentFileNode);

    if (pathParts.length === 0) {
      parentFileNode.children.push(this.getFileNodeFormDataFile(dataFile));
    } else {
      this.recursiveAddFolder(parentFileNode.children, dataFile, pathParts);
    }
  }

  private createLevelIfNotExist(parentFileNode: FileNode, firstFolderName: string, levelParent: FileNode[]): FileNode {
    if (parentFileNode === undefined) {
      parentFileNode = {
        name: firstFolderName,
      } as FileNode;
      levelParent.push(parentFileNode);
    }
    return parentFileNode;
  }

  private initChildrenAttributIfUndefined(parentFileNode: FileNode): void {
    if (parentFileNode.children === undefined) {
      parentFileNode.children = [];
    }
  }

  private addFileToRootDataFile(dataFile: DepositDataFile): void {
    this.fileTree.push(this.getFileNodeFormDataFile(dataFile));
  }

  private getFileNodeFormDataFile(dataFile: DepositDataFile): FileNode {
    return {
      name: dataFile.fileName,
      file: dataFile,
    } as FileNode;
  }

  applyTreeToDataSource(): void {
    this.dataSource.data = this.fileTree;
  }

  hasChild = (_: number, node: FileNode) => !!node.children && node.children.length > 0;

  delete(depositDataFile: DepositDataFile): void {
    this._deleteBS.next(depositDataFile);
  }

  resume(depositDataFile: DepositDataFile): void {
    this._resumeBS.next(depositDataFile);
  }

  refresh(): void {
    this._refreshBS.next();
  }

  download(depositDataFile: DepositDataFile): void {
    this._downloadBS.next(depositDataFile);
  }

  showHistory(depositDataFile: DepositDataFile): void {
    this._showHistoryBS.next(depositDataFile);
  }
}
