import {PersonExtended} from "@admin/models/person-extended.model";
import {
  ChangeDetectionStrategy,
  ChangeDetectorRef,
  Component,
  Input,
  OnInit,
  Output,
  ViewChild,
} from "@angular/core";
import {
  FormBuilder,
  Validators,
} from "@angular/forms";
import {MatSelectChange} from "@angular/material";
import {MatDialog} from "@angular/material/dialog";
import {DepositExtended} from "@app/features/deposit/models/deposits-extended.model";
import {OrganizationalUnitExtended} from "@app/features/deposit/models/organizational-unit-extended.model";
import {
  Deposit,
  Language,
  License,
  Person,
  PreservationPolicy,
  SubmissionPolicy,
} from "@app/generated-api";
import {SharedAbstractFormPresentational} from "@app/shared/components/presentationals/shared-abstract-form/shared-abstract-form.presentational";
import {LocalModelAttributeEnum} from "@app/shared/enums/model-attribute.enum";
import {BaseFormDefinition} from "@app/shared/models/base-form-definition.model";
import {sharedLicenseActionNameSpace} from "@app/shared/stores/license/shared-license.action";
import {sharedPersonActionNameSpace} from "@app/shared/stores/person/shared-person.action";
import {DepositPersonDialog} from "@deposit/components/dialogs/deposit-person/deposit-person.dialog";
import {SharedSearchableMultiSelectPresentational} from "@shared/components/presentationals/shared-searchable-multi-select/shared-searchable-multi-select.presentational";
import {BreakpointService} from "@shared/services/breakpoint.service";
import {SharedLicenseState} from "@shared/stores/license/shared-license.state";
import {SharedPersonState} from "@shared/stores/person/shared-person.state";
import _ from "lodash";
import {
  BehaviorSubject,
  Observable,
} from "rxjs";
import {
  distinctUntilChanged,
  skipWhile,
  takeWhile,
  tap,
} from "rxjs/operators";
import {
  DateUtil,
  isNullOrUndefined,
  ObservableUtil,
  PropertyName,
  ResourceNameSpace,
  SolidifyValidator,
} from "solidify-frontend";
import AccessEnum = Deposit.AccessEnum;

@Component({
  selector: "dlcm-deposit-form",
  templateUrl: "./deposit-form.presentational.html",
  styleUrls: ["./deposit-form.presentational.scss"],
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class DepositFormPresentational extends SharedAbstractFormPresentational<DepositExtended> implements OnInit {
  formDefinition: FormComponentFormDefinition = new FormComponentFormDefinition();

  accessEnumValues: string[] = [...Object.values(AccessEnum)];

  @Input()
  isReady: boolean = false;

  @Input()
  languages: Language[];

  @Input()
  listSubmissionPolicies: SubmissionPolicy[];

  @Input()
  listPreservationPolicies: PreservationPolicy[];

  @Input()
  organizationalUnits: OrganizationalUnitExtended[];

  @Input()
  selectedPersons: PersonExtended[];

  @Input()
  listPersons: PersonExtended[];

  @Input()
  personConnected: PersonExtended;

  @Input()
  selectedOrgUnit: Observable<OrganizationalUnitExtended>;

  @ViewChild("authorMultiSearchableSelect", {static: false})
  authorMultiSearchableSelect: SharedSearchableMultiSelectPresentational;

  readonly TIME_BEFORE_DISPLAY_TOOLTIP: number = 500;

  sharedPersonActionNameSpace: ResourceNameSpace = sharedPersonActionNameSpace;
  sharedPersonState: typeof SharedPersonState = SharedPersonState;

  sharedLicenseActionNameSpace: ResourceNameSpace = sharedLicenseActionNameSpace;
  sharedLicenseState: typeof SharedLicenseState = SharedLicenseState;
  licenceCallback: (value: License) => string = (value: License) => value.openLicenseId + " (" + value.title + ")";

  private readonly _orgUnitValueBS: BehaviorSubject<string | undefined> = new BehaviorSubject<string | undefined>(undefined);
  @Output("orgUnitChange")
  readonly orgUnitValueObs: Observable<string | undefined> = ObservableUtil.asObservable(this._orgUnitValueBS);

  constructor(protected readonly _changeDetectorRef: ChangeDetectorRef,
              private readonly _fb: FormBuilder,
              private readonly dialog: MatDialog,
              public readonly breakpointService: BreakpointService) {
    super(_changeDetectorRef);
  }

  ngOnInit(): void {
    super.ngOnInit();

    this.validationLicenseRequiredWhenAccessPublic();
  }

  private validationLicenseRequiredWhenAccessPublic(): void {
    const licenseFormControl = this.form.get(this.formDefinition.licenseId);
    this.subscribe(this.form.get(this.formDefinition.access).valueChanges.pipe(
      distinctUntilChanged(),
      tap(accessLevel => {
        if (accessLevel === AccessEnum.PUBLIC) {
          licenseFormControl.setValidators([Validators.required]);
        } else {
          licenseFormControl.setValidators([]);
        }
        licenseFormControl.updateValueAndValidity();
      }),
    ));
  }

  protected initNewForm(): void {
    this.form = this._fb.group({
      [this.formDefinition.organizationalUnitId]: ["", [Validators.required, SolidifyValidator]],
      [this.formDefinition.title]: ["", [Validators.required, SolidifyValidator]],
      [this.formDefinition.description]: ["", [Validators.required, SolidifyValidator]],
      [this.formDefinition.languageId]: [""],
      [this.formDefinition.publicationDate]: ["", [Validators.required, SolidifyValidator]],
      [this.formDefinition.collectionBegin]: [""],
      [this.formDefinition.collectionEnd]: [""],
      [this.formDefinition.access]: [AccessEnum.PUBLIC],
      [this.formDefinition.hasEmbargo]: [false],
      [this.formDefinition.licenseId]: ["", [Validators.required, SolidifyValidator]],
      [this.formDefinition.submissionPolicyId]: ["", [Validators.required, SolidifyValidator]],
      [this.formDefinition.preservationPolicyId]: ["", [Validators.required, SolidifyValidator]],
      [this.formDefinition.authors]: ["", [Validators.required, SolidifyValidator]],
    });
  }

  protected bindFormTo(deposit: DepositExtended): void {
    this.form = this._fb.group({
      [this.formDefinition.organizationalUnitId]: [deposit.organizationalUnitId, [Validators.required, SolidifyValidator]],
      [this.formDefinition.title]: [deposit.title, [Validators.required, SolidifyValidator]],
      [this.formDefinition.description]: [deposit.description, [Validators.required, SolidifyValidator]],
      [this.formDefinition.languageId]: [deposit.languageId],
      [this.formDefinition.publicationDate]: [deposit.publicationDate, [Validators.required, SolidifyValidator]],
      [this.formDefinition.collectionBegin]: [DateUtil.convertOffsetDateTimeIso8601ToDate(deposit.collectionBegin), [SolidifyValidator]],
      [this.formDefinition.collectionEnd]: [DateUtil.convertOffsetDateTimeIso8601ToDate(deposit.collectionEnd), [SolidifyValidator]],
      [this.formDefinition.access]: [deposit.access],
      [this.formDefinition.hasEmbargo]: [deposit.hasEmbargo],
      [this.formDefinition.licenseId]: [deposit.licenseId],
      [this.formDefinition.submissionPolicyId]: [deposit.submissionPolicyId, [Validators.required, SolidifyValidator]],
      [this.formDefinition.preservationPolicyId]: [deposit.preservationPolicyId, [Validators.required, SolidifyValidator]],
      [this.formDefinition.authors]: [_.map(this.selectedPersons, LocalModelAttributeEnum.resId), [Validators.required, SolidifyValidator]],
    });
    this.isValidWhenDisable = this.form.valid;
  }

  protected treatmentBeforeSubmit(deposit: DepositExtended): DepositExtended {
    deposit.publicationDate = DateUtil.convertToLocalDateDateSimple(deposit.publicationDate);
    deposit.collectionBegin = DateUtil.convertToOffsetDateTimeIso8601(deposit.collectionBegin);
    deposit.collectionEnd = DateUtil.convertToOffsetDateTimeIso8601(deposit.collectionEnd);
    return deposit;
  }

  addPeople(): void {
    this.subscribe(
      this.dialog.open(DepositPersonDialog, {
        width: "90%",
      }).afterClosed().pipe(
        tap((person: PersonExtended) => {
          if (isNullOrUndefined(person)) {
            return;
          }
          this.listPersons = [...this.listPersons, person];
          this.addAuthor(person);
        }),
      ),
    );
  }

  addMeAsAuthor(): void {
    this.addAuthor(this.personConnected);
  }

  addAuthor(person: Person): void {
    this.authorMultiSearchableSelect.add(person);
  }

  personConnectedAlreadyInAuthor(): boolean {
    return this.form.get(this.formDefinition.authors).value.includes(this.personConnected.resId);
  }

  orgUnitSelectionChange($event: MatSelectChange): void {
    this._orgUnitValueBS.next($event.value);
    let valueChange: boolean = false;
    this.subscribe(this.selectedOrgUnit.pipe(
      skipWhile(o => o === null || o === undefined),
      takeWhile(() => valueChange === false),
      tap(o => {
        if (o.defaultPreservationPolicy !== null) {
          this.form.get(this.formDefinition.preservationPolicyId).setValue(o.defaultPreservationPolicy.resId);
        }
        if (o.defaultSubmissionPolicy !== null) {
          this.form.get(this.formDefinition.submissionPolicyId).setValue(o.defaultSubmissionPolicy.resId);
        }
        valueChange = true;
      }),
    ));
  }

}

class FormComponentFormDefinition extends BaseFormDefinition {
  @PropertyName() organizationalUnitId: string;
  @PropertyName() title: string;
  @PropertyName() description: string;
  @PropertyName() languageId: string;
  @PropertyName() publicationDate: string;
  @PropertyName() collectionBegin: string;
  @PropertyName() collectionEnd: string;
  @PropertyName() authors: string;
  @PropertyName() access: string;
  @PropertyName() hasEmbargo: string;
  @PropertyName() licenseId: string;
  @PropertyName() submissionPolicyId: string;
  @PropertyName() preservationPolicyId: string;
}
