import {
  ChangeDetectionStrategy,
  ChangeDetectorRef,
  Component,
  OnInit,
  ViewChild,
} from "@angular/core";
import {MatDialog} from "@angular/material";
import {MatTabGroup} from "@angular/material/tabs";
import {
  ActivatedRoute,
  NavigationCancel,
  NavigationEnd,
  Router,
} from "@angular/router";
import {
  DepositDataFile,
  DepositDataFileStateEnum,
} from "@app/features/deposit/models/deposit-data-file.model";
import {DepositExtended} from "@app/features/deposit/models/deposits-extended.model";
import {DepositDataFileAction} from "@app/features/deposit/stores/data-file/deposit-data-file.action";
import {
  DepositAction,
  depositActionNameSpace,
} from "@app/features/deposit/stores/deposit.action";
import {
  DepositState,
  DepositStateModel,
} from "@app/features/deposit/stores/deposit.state";
import {DepositPersonAction} from "@app/features/deposit/stores/people/deposit-person.action";
import {Deposit} from "@app/generated-api";
import {SharedAbstractDetailRoutable} from "@app/shared/components/routables/shared-abstract-detail/shared-abstract-detail.routable";
import {LocalStateEnum} from "@app/shared/enums/local-state.enum";
import {LocalStateModel} from "@app/shared/models/local-state.model";
import {SharedOrgUnitAction} from "@app/shared/stores/organizational-unit/shared-organizational-unit.action";
import {AppState} from "@app/stores/app.state";
import {AppOrgUnitPersonRoleAction} from "@app/stores/organizational-unit-person-role/app-organizational-unit-person-role.action";
import {DepositFormPresentational} from "@deposit/components/presentationals/deposit-form/deposit-form.presentational";
import {DepositTabEnum} from "@deposit/enums/deposit-tab.enum";
import {DepositHelper} from "@deposit/helpers/deposit.helper";
import {DepositStatusHistoryAction} from "@deposit/stores/status-history/deposit-status-history.action";
import {Navigate} from "@ngxs/router-plugin";
import {
  Select,
  Store,
} from "@ngxs/store";
import {SharedHistoryDialog} from "@shared/components/dialogs/shared-history/shared-history.dialog";
import {ApplicationRoleEnum} from "@shared/enums/application-role.enum";
import {RoleEnum} from "@shared/enums/role.enum";
import {RoutesEnum} from "@shared/enums/routes.enum";
import {StatusHistoryDialog} from "@shared/models/status-history-dialog.model";
import {StatusHistory} from "@shared/models/status-history.model";
import {
  combineLatest,
  Observable,
} from "rxjs";
import {tap} from "rxjs/internal/operators/tap";
import {
  distinctUntilChanged,
  filter,
  skipWhile,
  take,
} from "rxjs/operators";
import {
  isEmptyArray,
  isFalsy,
  isNullOrUndefined,
  StoreUtil,
  TRANSLATE,
} from "solidify-frontend";

@Component({
  selector: "dlcm-deposit-detail-routable",
  templateUrl: "./deposit-detail.routable.html",
  styleUrls: ["./deposit-detail.routable.scss"],
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class DepositDetailRoutable extends SharedAbstractDetailRoutable<DepositExtended, DepositStateModel> implements OnInit {
  @Select(DepositState.isLoadingWithDependency) isLoadingWithDependencyObs: Observable<boolean>;
  @Select(DepositState.isReadyToBeDisplayed) isReadyToBeDisplayedObs: Observable<boolean>;
  @Select((state: LocalStateModel) => StoreUtil.isLoadingState(state.deposit.deposit_person)) isLoadingPersonObs: Observable<boolean>;
  @Select((state: LocalStateModel) => StoreUtil.isLoadingState(state.deposit.deposit_dataFile)) isLoadingDataFileObs: Observable<boolean>;
  @Select((state: LocalStateModel) => state.deposit.deposit_statusHistory.history) historyObs: Observable<StatusHistory[]>;
  @Select((state: LocalStateModel) => state.deposit.deposit_dataFile.list) listDataFileObs: Observable<DepositDataFile[]>;
  @Select(AppState.currentOrgUnitPersonRoleResId) currentPersonRoleObs: Observable<string[]>;
  @Select(AppState.currentUserApplicationRolesResId) currentUserApplicationRoleObs: Observable<string[]>;

  @ViewChild("formPresentational", {static: false})
  readonly formPresentational: DepositFormPresentational;

  @ViewChild("matTabGroup", {static: false})
  readonly matTabGroup: MatTabGroup;

  message: string;
  tabSelected: DepositTabEnum;
  canDoValidatorActions: boolean = false;
  canDoAlterationActions: boolean = false;

  readonly KEY_PARAM_NAME: string = "title";
  readonly KEY_DELETE_BUTTON: string = TRANSLATE("deposit.delete");
  readonly KEY_EDIT_BUTTON: string = TRANSLATE("deposit.edit");

  get statusEnum(): typeof Deposit.StatusEnum {
    return Deposit.StatusEnum;
  }

  constructor(protected store: Store,
              protected route: ActivatedRoute,
              protected router: Router,
              public dialog: MatDialog,
              private _changeDetector: ChangeDetectorRef) {
    super(store, route, dialog, LocalStateEnum.deposit, depositActionNameSpace);
  }

  ngOnInit(): void {
    super.ngOnInit();
    this.store.dispatch(new SharedOrgUnitAction.GetAll());

    this._computeCurrentTab();
    this.subscribe(this.router.events
      .pipe(
        filter(event => event instanceof NavigationCancel || event instanceof NavigationEnd),
        distinctUntilChanged(),
        tap(event => {
          this._computeCurrentTab();
          this.matTabGroup.selectedIndex = this.tabSelected;
        }),
      ),
    );

    this._computeCurrentUserRight();
  }

  private _computeCurrentTab(): void {
    const tabRouteSelected = DepositHelper.getTabRouteSelected(this.route);
    this.tabSelected = DepositHelper.getTabSelectedIndexWithRoute(tabRouteSelected);
  }

  private _computeCurrentUserRight(): void {
    const currentPerson = this.store.selectSnapshot(AppState.currentPerson);

    this.subscribe(this.currentObs.pipe(
      skipWhile(d => isNullOrUndefined(d)),
      take(1),
      tap(d => {
        this.store.dispatch(new AppOrgUnitPersonRoleAction.GetById(d.organizationalUnitId, currentPerson.resId));
      }),
    ));

    this.subscribe(
      combineLatest(this.currentPersonRoleObs, this.currentUserApplicationRoleObs)
        .pipe(
          distinctUntilChanged(),
          tap(([roles, applicationRoles]) => {
            this.canDoValidatorActions = this.canBeValidator(roles, applicationRoles);
            this.canDoAlterationActions = this.canDoAlteration(roles, applicationRoles);
            this._changeDetector.detectChanges();
          }),
        ),
    );
  }

  canBeValidator(roles: string[], applicationRoles: string[]): boolean {
    if ((!isNullOrUndefined(roles) && !isEmptyArray(roles))
      && (roles.includes(RoleEnum.APPROVER)
        || roles.includes(RoleEnum.STEWARD)
        || roles.includes(RoleEnum.MANAGER))) {
      return true;
    }
    return !isNullOrUndefined(applicationRoles)
      && (applicationRoles.includes(ApplicationRoleEnum.admin)
        || applicationRoles.includes(ApplicationRoleEnum.root));
  }

  canDoAlteration(roles: string[], applicationRoles: string[]): boolean {
    if ((!isNullOrUndefined(roles) && !isEmptyArray(roles))
      && (roles.includes(RoleEnum.CREATOR)
        || roles.includes(RoleEnum.APPROVER)
        || roles.includes(RoleEnum.STEWARD)
        || roles.includes(RoleEnum.MANAGER))) {
      return true;
    }
    return !isNullOrUndefined(applicationRoles)
      && (applicationRoles.includes(ApplicationRoleEnum.admin)
        || applicationRoles.includes(ApplicationRoleEnum.root));
  }

  canSubmit(list: DepositDataFile[]): boolean {
    if (isNullOrUndefined(list) || isEmptyArray(list)) {
      this.message = TRANSLATE("deposit.error.noFile");
      return false;
    }

    const result = !list.some(d => {
      if (d.status === DepositDataFileStateEnum.IN_ERROR) {
        this.message = TRANSLATE("deposit.error.fileInError");
        return true;
      }
      return false;
    });
    return result;
  }

  edit(): void {
    const tabRouteSelected = DepositHelper.getTabRouteSelected(this.route);
    this.store.dispatch(new Navigate([RoutesEnum.depositEdit, this._resId, tabRouteSelected]));
  }

  getSubResourceWithParentId(id: string): void {
    this.store.dispatch(new DepositDataFileAction.GetAll(id));
    this.store.dispatch(new DepositPersonAction.GetAll(id));
  }

  submit(): void {
    this.currentObs.pipe(
      take(1),
      tap(deposit => this.store.dispatch(new DepositAction.Submit(deposit))),
    ).subscribe();
  }

  showHistory(): void {
    this.store.dispatch(new DepositStatusHistoryAction.History(this._resId));
    const dialogRef = this.dialog.open(SharedHistoryDialog, {
      width: "800px",
      data: {
        resourceResId: this._resId,
        name: this.state,
        statusHistory: this.historyObs,
        isLoading: this.store.select((s: LocalStateModel) => s.deposit.deposit_statusHistory.isLoadingHistoryCounter > 0),
      } as StatusHistoryDialog,
    });
  }

  navigate(tabSelected: DepositTabEnum): void {
    const tabRouteSelected = DepositHelper.getTabRouteSelectedWithTabIndex(tabSelected);
    this.store.dispatch(new Navigate([RoutesEnum.depositDetail, this._resId, tabRouteSelected]));
  }

  approve(): void {
    this.currentObs.pipe(
      take(1),
      tap(deposit => this.store.dispatch(new DepositAction.Approve(deposit))),
    ).subscribe();
  }

  reject(): void {
    this.currentObs.pipe(
      take(1),
      tap(deposit => this.store.dispatch(new DepositAction.Reject(deposit))),
    ).subscribe();
  }

  backToEdit(): void {
    this.currentObs.pipe(
      take(1),
      tap(deposit => this.store.dispatch(new DepositAction.BackToEdit(deposit))),
    ).subscribe();
  }

  filesRetrieve(depositDataFiles: DepositDataFile[]): boolean {
    if (isFalsy(depositDataFiles)) {
      return false;
    }
    return true;
  }

  backToList(): void {
    this.store.dispatch(new Navigate([RoutesEnum.deposit]));
  }
}
