import {
  ChangeDetectionStrategy,
  Component,
  OnDestroy,
  OnInit,
  ViewChild,
} from "@angular/core";
import {MatDialog} from "@angular/material";
import {ActivatedRoute} from "@angular/router";
import {
  AccessOrganizationalUnit,
  Deposit,
  Language,
  Person,
  PreservationPolicy,
  SubmissionPolicy,
} from "@app/generated-api";
import {DepositFormPresentational} from "@deposit/components/presentationals/deposit-form/deposit-form.presentational";
import {DepositHelper} from "@deposit/helpers/deposit.helper";
import {DepositDataFile} from "@deposit/models/deposit-data-file.model";
import {DepositExtended} from "@deposit/models/deposits-extended.model";
import {OrganizationalUnitExtended} from "@deposit/models/organizational-unit-extended.model";
import {DepositDataFileAction} from "@deposit/stores/data-file/deposit-data-file.action";
import {depositActionNameSpace} from "@deposit/stores/deposit.action";
import {
  DepositState,
  DepositStateModel,
} from "@deposit/stores/deposit.state";
import {DepositPersonAction} from "@deposit/stores/people/deposit-person.action";
import {
  Select,
  Store,
} from "@ngxs/store";
import {SharedAbstractEditRoutable} from "@shared/components/routables/shared-abstract-edit/shared-abstract-edit.routable";
import {LocalStateEnum} from "@shared/enums/local-state.enum";
import {LocalStateModel} from "@shared/models/local-state.model";
import {StatusHistory} from "@shared/models/status-history.model";
import {SharedOrgUnitPreservationPolicyAction} from "@shared/stores/organizational-unit/preservation-policy/shared-organizational-unit-preservation-policy.action";
import {SharedOrgUnitAction} from "@shared/stores/organizational-unit/shared-organizational-unit.action";
import {SharedOrgUnitSubmissionPolicyAction} from "@shared/stores/organizational-unit/submission-policy/shared-organizational-unit-submission-policy.action";
import {Observable} from "rxjs";
import {
  skipWhile,
  take,
  tap,
} from "rxjs/operators";
import {TRANSLATE} from "solidify-frontend";

@Component({
  selector: "dlcm-deposit-metadata-routable",
  templateUrl: "./deposit-metadata.routable.html",
  //styleUrls: ["src/app/features/deposit/components/routables/deposit-detail/deposit-detail.routable.scss"],
  changeDetection: ChangeDetectionStrategy.OnPush,
})

export class DepositMetadataRoutable extends SharedAbstractEditRoutable<DepositExtended, DepositStateModel> implements OnInit, OnDestroy {
  @Select(DepositState.isLoadingWithDependency) isLoadingWithDependencyObs: Observable<boolean>;
  @Select(DepositState.isReadyToBeDisplayed) isReadyToBeDisplayedObs: Observable<boolean>;
  @Select((state: LocalStateModel) => state.deposit.deposit_statusHistory.history) historyObs: Observable<StatusHistory[]>;
  @Select((state: LocalStateModel) => state.application.application_person.current) personConnectedObs: Observable<Person>;
  @Select((state: LocalStateModel) => state.shared.shared_language.list) languagesObs: Observable<Language[]>;
  @Select((state: LocalStateModel) => state.shared.shared_person.list) listPersonObs: Observable<Person[]>;
  @Select((state: LocalStateModel) => state.application.application_authorizedOrganizationalUnit.list) organizationalUnitsObs: Observable<AccessOrganizationalUnit[]>;
  @Select((state: LocalStateModel) => state.deposit.deposit_dataFile.list) listDataFileObs: Observable<DepositDataFile[]>;
  @Select((state: LocalStateModel) => state.deposit.deposit_person.selected) selectedPersonObs: Observable<Person[]>;
  @Select((state: LocalStateModel) => state.deposit.current) currentDepositObs: Observable<Deposit>;
  @Select((state: LocalStateModel) => state.shared.shared_submissionPolicy.list) submissionPoliciesObs: Observable<SubmissionPolicy[]>;
  @Select((state: LocalStateModel) => state.shared.shared_preservationPolicy.list) preservationPoliciesObs: Observable<PreservationPolicy[]>;
  @Select((state: LocalStateModel) => state.shared.shared_organizationalUnit.shared_organizationalUnit_preservationPolicy.selected) listPreservationPoliciesObs: Observable<PreservationPolicy[]>;
  @Select((state: LocalStateModel) => state.shared.shared_organizationalUnit.shared_organizationalUnit_submissionPolicy.selected) listSubmissionPoliciesObs: Observable<SubmissionPolicy[]>;
  @Select((state: LocalStateModel) => state.shared.shared_organizationalUnit.current) selectedOrgUnitObs: Observable<OrganizationalUnitExtended>;

  @ViewChild("formPresentational", {static: false})
  readonly formPresentational: DepositFormPresentational;

  readonly KEY_PARAM_NAME: string = "title";
  readonly KEY_DELETE_BUTTON: string = TRANSLATE("deposit.delete");
  readonly KEY_EDIT_BUTTON: string = TRANSLATE("deposit.edit");

  public isReadonly: boolean = true;

  constructor(protected store: Store,
              protected route: ActivatedRoute,
              public dialog: MatDialog) {
    super(store, route, LocalStateEnum.deposit, depositActionNameSpace, null, {
      skipGetData: true,
      skipClearOnDestroy: true,
    });
  }

  ngOnInit(): void {
    super.ngOnInit();
    this.store.dispatch(new SharedOrgUnitAction.GetAll());
    this.isReadonly = DepositHelper.computeIsReadonly(this.route);
    this.initPolicies();
  }

  getSubResourceWithParentId(id: string): void {
    this.store.dispatch(new DepositDataFileAction.GetAll(id, undefined, true));
    this.store.dispatch(new DepositPersonAction.GetAll(id));
  }

  adaptPolicies(organizationalUnitId: string): void {
    this.store.dispatch(new SharedOrgUnitAction.GetById(organizationalUnitId));
    this.store.dispatch(new SharedOrgUnitPreservationPolicyAction.GetAll(organizationalUnitId));
    this.store.dispatch(new SharedOrgUnitSubmissionPolicyAction.GetAll(organizationalUnitId));
  }

  initPolicies(): void {
    this.subscribe(this.currentDepositObs.pipe(
      skipWhile((deposit: DepositExtended) => deposit === null || deposit === undefined),
      take(1),
      tap(deposit => {
        this.adaptPolicies(deposit.organizationalUnitId);
      })));
  }

  ngOnDestroy(): void {
    super.ngOnDestroy();
    this.store.dispatch(new SharedOrgUnitPreservationPolicyAction.Clear());
    this.store.dispatch(new SharedOrgUnitSubmissionPolicyAction.Clear());
  }
}
