import {
  HttpErrorResponse,
  HttpEventType,
} from "@angular/common/http";
import {FileUploadStatusEnum} from "@app/features/deposit/enums/file-upload-status.enum";
import {DataFileUploadHelper} from "@app/features/deposit/helpers/data-file-upload.helper";
import {DepositDataFile} from "@app/features/deposit/models/deposit-data-file.model";
import {DepositExtended} from "@app/features/deposit/models/deposits-extended.model";
import {UploadEventModel} from "@app/features/deposit/models/upload-event.model";
import {UploadFileStatus} from "@app/features/deposit/models/upload-file-status.model";
import {depositDataFileActionNameSpace} from "@app/features/deposit/stores/data-file/deposit-data-file.action";
import {
  DepositDataFileState,
  DepositDataFileStateModel,
} from "@app/features/deposit/stores/data-file/deposit-data-file.state";
import {
  DepositAction,
  depositActionNameSpace,
} from "@app/features/deposit/stores/deposit.action";
import {
  DepositPeopleState,
  DepositPersonStateModel,
} from "@app/features/deposit/stores/people/deposit-people.state";
import {DepositPersonAction} from "@app/features/deposit/stores/people/deposit-person.action";
import {Deposit} from "@app/generated-api";
import {ApiActionEnum} from "@app/shared/enums/api-action.enum";
import {PreIngestResourceApiEnum} from "@app/shared/enums/api.enum";
import {ErrorBackendKeyEnum} from "@app/shared/enums/error-backend-key.enum";
import {LocalStateEnum} from "@app/shared/enums/local-state.enum";
import {LocalModelAttributeEnum} from "@app/shared/enums/model-attribute.enum";
import {RoutesEnum} from "@app/shared/enums/routes.enum";
import {SharedLanguageAction} from "@app/shared/stores/language/shared-language.action";
import {SharedLicenseAction} from "@app/shared/stores/license/shared-license.action";
import {SharedPersonAction} from "@app/shared/stores/person/shared-person.action";
import {SharedPreservationPolicyAction} from "@app/shared/stores/preservation-policy/shared-preservation-policy.action";
import {SharedSubmissionPolicyAction} from "@app/shared/stores/submission-policy/shared-submission-policy.action";
import {AppAction} from "@app/stores/app.action";
import {
  DepositStatusHistoryState,
  DepositStatusHistoryStateModel,
} from "@deposit/stores/status-history/deposit-status-history.state";
import {Navigate} from "@ngxs/router-plugin";
import {
  Action,
  Actions,
  ofActionCompleted,
  Selector,
  State,
  StateContext,
  Store,
} from "@ngxs/store";
import {defaultStatusHistoryInitValue} from "@shared/stores/status-history/status-history.state";
import {Observable} from "rxjs";
import {
  catchError,
  map,
  tap,
} from "rxjs/operators";
import {
  ApiService,
  CompositionActionHelper,
  defaultAssociationRemoteStateInitValue,
  defaultCompositionStateInitValue,
  defaultResourceStateInitValue,
  ErrorDto,
  HttpStatus,
  isNullOrUndefined,
  isTruthy,
  NotificationService,
  ObjectUtil,
  OverrideDefaultAction,
  ResourceState,
  ResourceStateModel,
  StoreUtil,
  TRANSLATE,
} from "solidify-frontend";
import {environment} from "../../../../environments/environment";

export interface DepositStateModel extends ResourceStateModel<DepositExtended> {
  deposit_dataFile: DepositDataFileStateModel;
  deposit_person: DepositPersonStateModel;
  isLoadingDataFile: boolean;
  uploadStatus: UploadFileStatus[];
  deposit_statusHistory: DepositStatusHistoryStateModel;
}

@State<DepositStateModel>({
  name: LocalStateEnum.deposit,
  defaults: {
    ...defaultResourceStateInitValue,
    deposit_dataFile: {...defaultCompositionStateInitValue, deposit_dataFile_statusHistory: {...defaultStatusHistoryInitValue}},
    deposit_person: {...defaultAssociationRemoteStateInitValue},
    isLoadingDataFile: false,
    uploadStatus: [],
    deposit_statusHistory: {...defaultStatusHistoryInitValue},
  },
  children: [
    DepositDataFileState,
    DepositPeopleState,
    DepositStatusHistoryState,
  ],
})
export class DepositState extends ResourceState<DepositExtended> {
  private readonly _FILE_KEY: string = "file";
  private readonly _CATEGORY_KEY: string = "category";
  private readonly _TYPE_KEY: string = "type";
  private readonly _FOLDER_KEY: string = "folder";

  constructor(protected apiService: ApiService,
              protected store: Store,
              protected notificationService: NotificationService,
              protected actions$: Actions) {
    super(apiService, store, notificationService, actions$, {
      nameSpace: depositActionNameSpace,
      routeRedirectUrlAfterSuccessAction: RoutesEnum.deposit,
      notificationResourceCreateTextToTranslate: TRANSLATE("deposit.notification.resource.create"),
      notificationResourceDeleteTextToTranslate: TRANSLATE("deposit.notification.resource.delete"),
      notificationResourceUpdateTextToTranslate: TRANSLATE("deposit.notification.resource.update"),
    });
  }

  protected get _urlResource(): string {
    return PreIngestResourceApiEnum.deposits;
  }

  @Selector()
  static isLoading(state: DepositStateModel): boolean {
    return StoreUtil.isLoadingState(state);
  }

  @Selector()
  static isLoadingWithDependency(state: DepositStateModel): boolean {
    return this.isLoading(state)
      || StoreUtil.isLoadingState(state.deposit_person)
      || StoreUtil.isLoadingState(state.deposit_dataFile);
  }

  @Selector()
  static isReadyToBeDisplayed(state: DepositStateModel): boolean {
    return this.isReadyToBeDisplayedInCreateMode
      && !isNullOrUndefined(state.current)
      && !isNullOrUndefined(state.deposit_dataFile.list)
      && !isNullOrUndefined(state.deposit_person.selected);
  }

  @Selector()
  static isReadyToBeDisplayedInCreateMode(state: DepositStateModel): boolean {
    return true;
  }

  @Selector()
  static current(state: DepositStateModel): DepositExtended {
    return state.current;
  }

  @Action(DepositAction.LoadResource)
  loadResource(ctx: StateContext<DepositStateModel>, action: DepositAction.LoadResource): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
    });

    StoreUtil.dispatchParallelActionAndWaitForSubActionsCompletion(ctx, [
      {
        action: new SharedLanguageAction.GetAll(),
        subActionCompletions: [
          this.actions$.pipe(ofActionCompleted(SharedLanguageAction.GetAllSuccess)),
          this.actions$.pipe(ofActionCompleted(SharedLanguageAction.GetAllFail)),
        ],
      },
      {
        action: new SharedLicenseAction.GetAll(),
        subActionCompletions: [
          this.actions$.pipe(ofActionCompleted(SharedLicenseAction.GetAllSuccess)),
          this.actions$.pipe(ofActionCompleted(SharedLicenseAction.GetAllFail)),
        ],
      },
      {
        action: new SharedPreservationPolicyAction.GetAll(),
        subActionCompletions: [
          this.actions$.pipe(ofActionCompleted(SharedPreservationPolicyAction.GetAllSuccess)),
          this.actions$.pipe(ofActionCompleted(SharedPreservationPolicyAction.GetAllFail)),
        ],
      },
      {
        action: new SharedSubmissionPolicyAction.GetAll(),
        subActionCompletions: [
          this.actions$.pipe(ofActionCompleted(SharedSubmissionPolicyAction.GetAllSuccess)),
          this.actions$.pipe(ofActionCompleted(SharedSubmissionPolicyAction.GetAllFail)),
        ],
      },
      {
        action: new SharedPersonAction.GetAll(),
        subActionCompletions: [
          this.actions$.pipe(ofActionCompleted(SharedPersonAction.GetAllSuccess)),
          this.actions$.pipe(ofActionCompleted(SharedPersonAction.GetAllFail)),
        ],
      },
    ]).subscribe(success => {
      if (success) {
        ctx.dispatch(new DepositAction.LoadResourceSuccess(action));
      } else {
        ctx.dispatch(new DepositAction.LoadResourceFail(action));
      }
    });
  }

  @OverrideDefaultAction()
  @Action(DepositAction.Create)
  create(ctx: StateContext<DepositStateModel>, action: DepositAction.Create): Observable<DepositExtended> {
    return super.internalCreate(ctx, action)
      .pipe(
        tap(model => {
          this.updateSubResource(model, action, ctx)
            .subscribe(success => {
              if (success) {
                ctx.dispatch(new DepositAction.CreateSuccess(action, model));
              } else {
                ctx.dispatch(new DepositAction.CreateFail(action));
              }
            });
        }),
      );
  }

  @OverrideDefaultAction()
  @Action(DepositAction.Update)
  update(ctx: StateContext<DepositStateModel>, action: DepositAction.Update): Observable<DepositExtended> {
    return super.internalUpdate(ctx, action)
      .pipe(
        tap(model => {
          this.updateSubResource(model, action, ctx)
            .subscribe(success => {
              if (success) {
                ctx.dispatch(new DepositAction.UpdateSuccess(action, model));
              } else {
                ctx.dispatch(new DepositAction.UpdateFail(action));
              }
            });
        }),
      );
  }

  private updateSubResource(model: DepositExtended, action: DepositAction.Create | DepositAction.Update, ctx: StateContext<DepositStateModel>): Observable<boolean> {
    const depositId = model.resId;
    const newAuthors: string[] = action.modelFormControlEvent.model[LocalModelAttributeEnum.authors];

    return StoreUtil.dispatchParallelActionAndWaitForSubActionsCompletion(ctx, [{
      action: new DepositPersonAction.Update(depositId, newAuthors),
      subActionCompletions: [
        this.actions$.pipe(ofActionCompleted(DepositPersonAction.UpdateSuccess)),
        this.actions$.pipe(ofActionCompleted(DepositPersonAction.UpdateFail)),
      ],
    }]);
  }

  @Action(DepositAction.Submit)
  submit(ctx: StateContext<DepositStateModel>, action: DepositAction.Submit): Observable<DepositExtended> {
    let submissionWithApproval = false;
    let submissionPolicy = action.deposit.submissionPolicy;
    if (isNullOrUndefined(submissionPolicy)) {
      submissionPolicy = action.deposit.organizationalUnit.defaultSubmissionPolicy;
    }
    if (!isNullOrUndefined(submissionPolicy)) {
      submissionWithApproval = submissionPolicy.submissionApproval;
    }

    const updateDeposit = {} as DepositExtended;
    if (isTruthy(submissionWithApproval)) {
      updateDeposit.status = Deposit.StatusEnum.INVALIDATION;
    } else {
      updateDeposit.status = Deposit.StatusEnum.APPROVED;
    }
    return this.apiService.patchById(this._urlResource, action.deposit.resId, updateDeposit).pipe(
      tap(deposit => ctx.dispatch(new DepositAction.SubmitSuccess(deposit))),
      catchError(error => {
        ctx.dispatch(new DepositAction.SubmitFail());
        throw error;
      }),
    );
  }

  @Action(DepositAction.SubmitSuccess)
  submitSuccess(ctx: StateContext<DepositStateModel>, action: DepositAction.SubmitSuccess): void {
    ctx.dispatch(new Navigate([RoutesEnum.deposit]));
    this.notificationService.showSuccess(TRANSLATE("deposit.notification.submit.success"), true);
  }

  @Action(DepositAction.SubmitFail)
  submitFail(ctx: StateContext<DepositStateModel>, action: DepositAction.SubmitFail): void {
    this.notificationService.showError(TRANSLATE("deposit.notification.submit.fail"), true);
  }

  @Action(DepositAction.Approve)
  approve(ctx: StateContext<DepositStateModel>, action: DepositAction.Approve): Observable<DepositExtended> {
    const updateDeposit = {} as DepositExtended;
    if (isTruthy(Deposit.StatusEnum.INVALIDATION)) {
      updateDeposit.status = Deposit.StatusEnum.APPROVED;
    }
    return this.apiService.patchById(this._urlResource, action.deposit.resId, updateDeposit).pipe(
      tap(deposit => ctx.dispatch(new DepositAction.ApproveSuccess(deposit))),
      catchError(error => {
        ctx.dispatch(new DepositAction.ApproveFail());
        throw error;
      }),
    );
  }

  @Action(DepositAction.ApproveSuccess)
  approveSuccess(ctx: StateContext<DepositStateModel>, action: DepositAction.ApproveSuccess): void {
    ctx.dispatch(new Navigate([RoutesEnum.deposit]));
    this.notificationService.showSuccess(TRANSLATE("deposit.notification.approve.success"), true);
  }

  @Action(DepositAction.ApproveFail)
  approveFail(ctx: StateContext<DepositStateModel>, action: DepositAction.ApproveFail): void {
    this.notificationService.showError(TRANSLATE("deposit.notification.approve.fail"), true);
  }

  @Action(DepositAction.Reject)
  reject(ctx: StateContext<DepositStateModel>, action: DepositAction.Reject): Observable<DepositExtended> {
    const updateDeposit = {} as DepositExtended;
    if (isTruthy(Deposit.StatusEnum.INVALIDATION)) {
      updateDeposit.status = Deposit.StatusEnum.REJECTED;
    }
    return this.apiService.patchById(this._urlResource, action.deposit.resId, updateDeposit).pipe(
      tap(deposit => ctx.dispatch(new DepositAction.RejectSuccess(deposit))),
      catchError(error => {
        ctx.dispatch(new DepositAction.RejectFail());
        throw error;
      }),
    );
  }

  @Action(DepositAction.RejectSuccess)
  rejectSuccess(ctx: StateContext<DepositStateModel>, action: DepositAction.RejectSuccess): void {
    ctx.dispatch(new Navigate([RoutesEnum.deposit]));
    this.notificationService.showSuccess(TRANSLATE("deposit.notification.reject.success"), true);
  }

  @Action(DepositAction.RejectFail)
  rejectFail(ctx: StateContext<DepositStateModel>, action: DepositAction.RejectFail): void {
    this.notificationService.showError(TRANSLATE("deposit.notification.reject.fail"), true);
  }

  @Action(DepositAction.BackToEdit)
  backToEdit(ctx: StateContext<DepositStateModel>, action: DepositAction.BackToEdit): Observable<DepositExtended> {
    const updateDeposit = {} as DepositExtended;
    if (isTruthy(Deposit.StatusEnum.INVALIDATION)) {
      updateDeposit.status = Deposit.StatusEnum.INPROGRESS;
    }
    return this.apiService.patchById(this._urlResource, action.deposit.resId, updateDeposit).pipe(
      tap(deposit => ctx.dispatch(new DepositAction.BackToEditSuccess(deposit))),
      catchError(error => {
        ctx.dispatch(new DepositAction.BackToEditFail());
        throw error;
      }),
    );
  }

  @Action(DepositAction.BackToEditSuccess)
  backToEditSuccess(ctx: StateContext<DepositStateModel>, action: DepositAction.BackToEditSuccess): void {
    ctx.dispatch(new Navigate([RoutesEnum.deposit]));
    this.notificationService.showSuccess(TRANSLATE("deposit.notification.backToEdit.success"), true);
  }

  @Action(DepositAction.BackToEditFail)
  backToEditFail(ctx: StateContext<DepositStateModel>, action: DepositAction.BackToEditFail): void {
    this.notificationService.showError(TRANSLATE("deposit.notification.backToEdit.fail"), true);
  }

  @Action(DepositAction.SendDataFile)
  sendDataFile(ctx: StateContext<DepositStateModel>, action: DepositAction.SendDataFile): Observable<void> {
    const formData = new FormData();
    let apiAction = ApiActionEnum.UL;
    if (action.isArchive) {
      apiAction = ApiActionEnum.UL_ARCHIVE;
    } else {
      formData.append(this._FOLDER_KEY, action.fileUploadWrapper.subDirectory);
    }
    formData.append(this._FILE_KEY, action.fileUploadWrapper.file, action.fileUploadWrapper.file.name);
    formData.append(this._CATEGORY_KEY, action.fileUploadWrapper.dataCategory);
    formData.append(this._TYPE_KEY, action.fileUploadWrapper.dataType);
    const uploadFileStatus = DataFileUploadHelper.addToUploadStatus(ctx, action.fileUploadWrapper);
    ctx.dispatch(new AppAction.PreventExit(TRANSLATE("notification.uploadInProgress")));

    return this.apiService.upload<DepositDataFile>(`${PreIngestResourceApiEnum.deposits}/${action.parentId}/${apiAction}`, formData)
      .pipe(
        map((event: UploadEventModel) => {
          switch (event.type) {
            case HttpEventType.UploadProgress:
              const currentUploadFileStatus = DataFileUploadHelper.getCurrentUploadFileStatus(ctx, uploadFileStatus.guid);
              if (currentUploadFileStatus.status === FileUploadStatusEnum.canceled) {
                ctx.dispatch(new DepositAction.CancelDataFileSending(action.parentId, uploadFileStatus));
                throw environment.errorToSkipInErrorHandler;
              }
              DataFileUploadHelper.updateInProgressUploadFileStatus(ctx, currentUploadFileStatus, event);
              return;
            case HttpEventType.Response:
              if (event.status === HttpStatus.OK) {
                ctx.dispatch(new DepositAction.SendDataFileSuccess(action.parentId, uploadFileStatus, event.body));
              } else {
                ctx.dispatch(new DepositAction.SendDataFileFail(uploadFileStatus, undefined));
              }
              return;
            default:
              return;
          }
        }),
        catchError((error: Error | HttpErrorResponse) => {
          if (error === environment.errorToSkipInErrorHandler) {
            throw error;
          }
          let errorDto: ErrorDto = undefined;
          if (error instanceof HttpErrorResponse) {
            errorDto = error.error;
          }
          ctx.dispatch(new DepositAction.SendDataFileFail(uploadFileStatus, errorDto));
          throw error;
        }),
      );
  }

  @Action(DepositAction.SendDataFileSuccess)
  sendDataFileSuccess(ctx: StateContext<DepositStateModel>, action: DepositAction.SendDataFileSuccess): void {
    this.notificationService.showSuccess(TRANSLATE("deposit.file.upload.notification.success"), true);
    ctx.dispatch(CompositionActionHelper.getAll(depositDataFileActionNameSpace, action.parentId, undefined, true));
    const uploadFileStatusCopied = ObjectUtil.clone(action.uploadFileStatus);
    DataFileUploadHelper.updateCompletedUploadFileStatus(ctx, uploadFileStatusCopied, action.depositDataFile);
    this.checkPendingDownload(ctx);
  }

  @Action(DepositAction.SendDataFileFail)
  sendDataFileFail(ctx: StateContext<DepositStateModel>, action: DepositAction.SendDataFileFail): void {
    let errorMessage = TRANSLATE("deposit.file.upload.notification.error");
    if (action.errorDto.message === ErrorBackendKeyEnum.UPLOAD_DUPLICATE_DATA_FILES) {
      errorMessage = TRANSLATE("error.upload.duplicateDataFiles");
    }
    this.notificationService.showError(errorMessage, true);
    const uploadFileStatusCopied = ObjectUtil.clone(action.uploadFileStatus);
    DataFileUploadHelper.updateErrorUploadFileStatus(ctx, uploadFileStatusCopied, errorMessage);
    this.checkPendingDownload(ctx);
  }

  private checkPendingDownload(ctx: StateContext<DepositStateModel>): void {
    if (ctx.getState()
      .uploadStatus
      .filter(u => u.status === FileUploadStatusEnum.inProgress || u.status === FileUploadStatusEnum.started).length === 0) {
      ctx.dispatch(new AppAction.CancelPreventExit());
    }
  }

  @Action(DepositAction.RetrySendDataFile)
  retrySendDataFile(ctx: StateContext<DepositStateModel>, action: DepositAction.RetrySendDataFile): void {
    DataFileUploadHelper.removeToUploadStatus(ctx, action.uploadFileStatus);
    ctx.dispatch(new DepositAction.SendDataFile(action.parentId, action.uploadFileStatus.fileUploadWrapper));
  }

  @Action(DepositAction.MarkAsCancelDataFileSending)
  markAsCancelDataFileSending(ctx: StateContext<DepositStateModel>, action: DepositAction.MarkAsCancelDataFileSending): void {
    const uploadFileStatus = DataFileUploadHelper.getCurrentUploadFileStatus(ctx, action.uploadFileStatus.guid);
    if (uploadFileStatus.status === FileUploadStatusEnum.inProgress) {
      DataFileUploadHelper.updateCancelUploadFileStatus(ctx, uploadFileStatus);
    } else {
      ctx.dispatch(new DepositAction.CancelDataFileSending(action.parentId, uploadFileStatus));
    }
  }

  @Action(DepositAction.CancelDataFileSending)
  cancelDataFileSending(ctx: StateContext<DepositStateModel>, action: DepositAction.MarkAsCancelDataFileSending): void {
    DataFileUploadHelper.removeToUploadStatus(ctx, action.uploadFileStatus);
    this.notificationService.showInformation(TRANSLATE("deposit.file.upload.notification.info.canceled"), true);
    this.checkPendingDownload(ctx);
  }
}
