import {
  ChangeDetectionStrategy,
  Component,
  OnInit,
} from "@angular/core";
import {SearchScopeEnum} from "@app/features/home/enums/search-scope.enum";
import {Archive} from "@app/features/home/models/archive.model";
import {HomeAction} from "@app/features/home/stores/home.action";
import {HomeState} from "@app/features/home/stores/home.state";
import {HomeOrgUnitAction} from "@app/features/home/stores/organizational-unit/home-organizational-unit.action";
import {AccessOrganizationalUnit} from "@app/generated-api";
import {Deposit} from "@app/generated-api/model/deposit.model";
import {SharedAbstractPresentational} from "@app/shared/components/presentationals/shared-abstract/shared-abstract.presentational";
import {FieldTypeEnum} from "@app/shared/enums/field-type.enum";
import {RoutesEnum} from "@app/shared/enums/routes.enum";
import {ArchiveMetadata} from "@app/shared/models/business/archive-metadata.model";
import {DataTableColumns} from "@app/shared/models/data-table-columns.model";
import {LocalStateModel} from "@app/shared/models/local-state.model";
import {MetadataUtil} from "@app/shared/utils/metadata.util";
import {TranslateService} from "@ngx-translate/core";
import {Navigate} from "@ngxs/router-plugin";
import {
  Select,
  Store,
} from "@ngxs/store";
import {
  BehaviorSubject,
  combineLatest,
  Observable,
} from "rxjs";
import {filter} from "rxjs/internal/operators/filter";
import {tap} from "rxjs/internal/operators/tap";
import {
  isNotNullNorUndefined,
  QueryParameters,
  StoreUtil,
  TRANSLATE,
} from "solidify-frontend";

@Component({
  selector: "dlcm-home-search-routable",
  templateUrl: "./home-search.routable.html",
  styleUrls: ["./home-search.routable.scss"],
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class HomeSearchRoutable extends SharedAbstractPresentational implements OnInit {
  @Select((state: LocalStateModel) => state.application.isLoggedIn) isLoggedInObs: Observable<boolean>;
  @Select((state: LocalStateModel) => state.home.search) searchObs: Observable<string>;
  @Select((state: LocalStateModel) => state.home.rss) rssObs: Observable<boolean>;
  @Select(HomeState.isLoading) isLoadingObs: Observable<boolean>;
  @Select(HomeState.list) listObs: Observable<ArchiveMetadata[]>;
  @Select(HomeState.searchScope) searchScopeObs: Observable<SearchScopeEnum>;
  @Select((state: LocalStateModel) => state.home.queryParameters) queryParametersObs: Observable<ArchiveMetadata[]>;
  @Select((state: LocalStateModel) => state.home.home_organizationalUnit.list) listOrganizationalUnitObs: Observable<AccessOrganizationalUnit[]>;
  @Select((state: LocalStateModel) => StoreUtil.isLoadingState(state.home.home_organizationalUnit)) isLoadingOrgUnitObs: Observable<boolean>;

  columns: DataTableColumns[];

  listAdaptedBS: BehaviorSubject<Archive[]> = new BehaviorSubject([]);
  listAdaptedObs: Observable<Archive[]> = this.listAdaptedBS.asObservable();

  constructor(private store: Store,
              private translate: TranslateService) {
    super();
  }

  ngOnInit(): void {
    super.ngOnInit();
    this.store.dispatch(new HomeOrgUnitAction.GetDefaultMaximumOrgUnit());
    this.store.dispatch(new HomeAction.GetRss());

    this.columns = [
      {
        field: "organizationalUnit",
        header: TRANSLATE("access.table.header.organizationalUnit"),
        type: FieldTypeEnum.string,
        order: 0,
        isSortable: false,
        isFilterable: false,
      },
      {
        field: "title",
        header: TRANSLATE("access.table.header.title"),
        type: FieldTypeEnum.string,
        order: 0,
        isSortable: false,
        isFilterable: false,
      },
      {
        field: "yearPublicationDate",
        header: TRANSLATE("access.table.header.yearPublicationDate"),
        type: FieldTypeEnum.number,
        order: 0,
        isSortable: false,
        isFilterable: false,
      },
      {
        field: "accessLevel",
        header: TRANSLATE("access.table.header.accessLevel"),
        type: FieldTypeEnum.string,
        order: 0,
        isSortable: false,
        isFilterable: false,
      },
    ];

    this.subscribe(
      combineLatest(this.listObs, this.listOrganizationalUnitObs, this.searchScopeObs)
        .pipe(
          filter(([list, listOrgUnits]) => isNotNullNorUndefined(list) && isNotNullNorUndefined(listOrgUnits)),
          tap(([listArchives, listOrgUnits, searchScope]) => {
            const newList: Archive[] = [];
            listArchives.forEach((archive: ArchiveMetadata) => {
              newList.push({
                resId: archive.resId,
                organizationalUnit: this.getOrganizationalUnitNameFromId(
                  MetadataUtil.getOrganizationalUnitResId(archive.metadata),
                  listOrgUnits,
                ),
                title: MetadataUtil.getTitle(searchScope, archive.metadata),
                yearPublicationDate: MetadataUtil.getPublicationYear(searchScope, archive.metadata),
                accessLevel: MetadataUtil.getAccessLevel(searchScope, archive.metadata) as Deposit.AccessEnum,
              });
            });
            this.listAdaptedBS.next(newList);
          }),
        ),
    );
  }

  getOrganizationalUnitNameFromId(resId: string, listOrgUnits: AccessOrganizationalUnit[]): string {
    const orgUnits = listOrgUnits.find(o => o.resId === resId);
    if (orgUnits) {
      return orgUnits.name;
    }
    return this.translate.instant("access.organizationalUnitNotFound");
  }

  search(searchTerm: string, queryParameters?: QueryParameters): void {
    this.store.dispatch(new HomeAction.Search(true, searchTerm, queryParameters));
  }

  onQueryParametersEvent($event: QueryParameters): void {
    this.store.dispatch(new HomeAction.ChangeQueryParameters($event));
  }

  showDetail(archive: Archive): void {
    this.store.dispatch(new Navigate([RoutesEnum.homePageDetail, archive.resId]));
  }

  searchScopeChange(searchScope: SearchScopeEnum): void {
    this.store.dispatch(new HomeAction.ChangeSearchScope(searchScope));
  }

  goToArchivalBrowsing(): void {
    this.store.dispatch(new Navigate([RoutesEnum.homePageBrowsing]));
  }
}
