import {
  HttpClient,
  HttpHeaders,
  HttpResponse,
} from "@angular/common/http";
import {SearchScopeEnum} from "@app/features/home/enums/search-scope.enum";
import {Rss} from "@app/features/home/models/rss.model";
import {HomeAction} from "@app/features/home/stores/home.action";
import {
  HomeOrganizationalUnitState,
  HomeOrganizationalUnitStateModel,
} from "@app/features/home/stores/organizational-unit/home-organizational-unit.state";
import {Order} from "@app/generated-api/model/order.model";
import {AccessResourceApiEnum} from "@app/shared/enums/api.enum";
import {LocalStateEnum} from "@app/shared/enums/local-state.enum";
import {urlSeparator} from "@app/shared/enums/routes.enum";
import {SessionStorageEnum} from "@app/shared/enums/session-storage.enum";
import {ArchiveMetadata} from "@app/shared/models/business/archive-metadata.model";
import {UserPreferencesUtil} from "@app/shared/utils/user-preferences.util";
import {AppStateModel} from "@app/stores/app.state";
import {
  Action,
  Actions,
  Selector,
  State,
  StateContext,
  Store,
} from "@ngxs/store";
import {ApiActionEnum} from "@shared/enums/api-action.enum";
import {saveAs} from "file-saver";
import {NgxXml2jsonService} from "ngx-xml2json";
import {
  interval,
  Observable,
  of,
} from "rxjs";
import {tap} from "rxjs/internal/operators/tap";
import {
  catchError,
  map,
  mergeMap,
  startWith,
  takeWhile,
} from "rxjs/operators";
import {
  ApiService,
  CollectionTyped,
  HttpStatus,
  isNullOrUndefined,
  isWhiteString,
  MapUtil,
  NotificationService,
  QueryParameters,
  QueryParametersUtil,
  ResourceStateModel,
  StoreUtil,
  StringUtil,
  TRANSLATE,
} from "solidify-frontend";
import {environment} from "../../../../environments/environment";
import DownloadFail = HomeAction.DownloadFail;

export interface HomeStateModel extends ResourceStateModel<ArchiveMetadata> {
  search: string;
  searchScope: SearchScopeEnum;
  rss: Rss | undefined;
  home_organizationalUnit: HomeOrganizationalUnitStateModel;
  isLoadingCounterDownloadPreparation: number;
}

@State<HomeStateModel>({
  name: LocalStateEnum.home,
  defaults: {
    searchScope: UserPreferencesUtil.getPreferredHomePageSearchScope(),
    search: null,
    total: 0,
    isLoadingCounter: 0,
    list: [],
    current: null,
    queryParameters: new QueryParameters(environment.defaultPageSizeHomePage),
    rss: undefined,
    home_organizationalUnit: {isLoadingCounter: 0} as HomeOrganizationalUnitStateModel,
    isLoadingCounterDownloadPreparation: 0,
  },
  children: [
    HomeOrganizationalUnitState,
  ],
})
export class HomeState {
  constructor(private apiService: ApiService,
              protected store: Store,
              protected httpClient: HttpClient,
              protected notificationService: NotificationService,
              private ngxXml2jsonService: NgxXml2jsonService,
              protected actions$: Actions) {
  }

  private readonly DEFAULT_SEARCH: string = "*";

  @Selector()
  static list(state: HomeStateModel): ArchiveMetadata[] {
    return state.list;
  }

  @Selector()
  static current(state: HomeStateModel): ArchiveMetadata {
    return state.current;
  }

  @Selector()
  static searchScope(state: HomeStateModel): SearchScopeEnum {
    return state.searchScope;
  }

  @Selector()
  static isLoading(state: HomeStateModel): boolean {
    return StoreUtil.isLoadingState(state);
  }

  @Selector()
  static isLoadingPrepareDownload(state: HomeStateModel): boolean {
    return state.isLoadingCounterDownloadPreparation > 0;
  }

  @Action(HomeAction.ChangeQueryParameters)
  changeQueryParameters(ctx: StateContext<HomeStateModel>, action: HomeAction.ChangeQueryParameters): void {
    ctx.patchState({
      queryParameters: action.queryParameters,
    });
    ctx.dispatch(new HomeAction.Search(false, ctx.getState().search));
  }

  @Action(HomeAction.ChangeSearchScope)
  changeSearchScope(ctx: StateContext<HomeStateModel>, action: HomeAction.ChangeSearchScope): void {
    ctx.patchState({
      searchScope: action.searchScope,
    });
    sessionStorage.setItem(SessionStorageEnum.homePageSearchScope, action.searchScope);
    const search = ctx.getState().search;
    if (!isWhiteString(search) && !isNullOrUndefined(search)) {
      ctx.dispatch(new HomeAction.Search(true, ctx.getState().search));
    }
  }

  @Action(HomeAction.Search)
  search(ctx: StateContext<HomeStateModel>, action: HomeAction.Search): Observable<CollectionTyped<ArchiveMetadata>> {
    const searchString = this.getSearchStringToApply(ctx, action);
    let queryParameters = StoreUtil.getQueryParametersToApply(action.queryParameters, ctx);
    if (action.resetPagination) {
      queryParameters = QueryParametersUtil.resetToFirstPage(queryParameters);
    }
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
      search: searchString,
      queryParameters: queryParameters,
    });

    const pathFilter = "query";
    MapUtil.addOrUpdate(queryParameters.search.searchItems, pathFilter, searchString);

    if (searchString === StringUtil.stringEmpty) {
      ctx.dispatch(new HomeAction.SearchSuccess(null));
    } else {
      return this.apiService.get<ArchiveMetadata>(this.getUrlMetadata(ctx) + urlSeparator + "search", queryParameters)
        .pipe(
          StoreUtil.cancelUncompleted(ctx, this.actions$, [HomeAction.Search]),
          tap(collection => ctx.dispatch(new HomeAction.SearchSuccess(collection))),
          catchError(error => {
            ctx.dispatch(new HomeAction.SearchFail());
            throw error;
          }),
        );
    }
  }

  private getUrlMetadata(ctx: StateContext<HomeStateModel>): string {
    const isLoggedIn: boolean = this.store.selectSnapshot<boolean>(state => (state.application as AppStateModel).isLoggedIn);
    if (!isLoggedIn) {
      return AccessResourceApiEnum.publicMetadata;
    }
    switch (ctx.getState().searchScope) {
      case SearchScopeEnum.all:
      // return AccessResourceApiEnum.allMetadata; // TO DISCUSS : All metadata endpoint display public + private metadata = we have all entry duplicate in datatable
      case SearchScopeEnum.private:
        return AccessResourceApiEnum.privateMetadata;
      case SearchScopeEnum.public:
      default:
        return AccessResourceApiEnum.publicMetadata;
    }
  }

  private getSearchStringToApply(ctx: StateContext<HomeStateModel>, action: HomeAction.Search): string {
    if (isNullOrUndefined(action.search)) {
      return this.getDefaultSearchStringIfNotDefined(ctx);
    }
    return action.search;
  }

  private getDefaultSearchStringIfNotDefined(ctx: StateContext<HomeStateModel>): string {
    if (ctx.getState().search === null || ctx.getState().search === undefined) {
      return this.DEFAULT_SEARCH;
    }
    return ctx.getState().search;
  }

  @Action(HomeAction.SearchSuccess)
  searchDipSuccess(ctx: StateContext<HomeStateModel>, action: HomeAction.SearchSuccess): void {
    const queryParameters = StoreUtil.updateQueryParameters(ctx, action.list);

    ctx.patchState({
      list: action.list === null || action.list === undefined ? [] : action.list._data,
      total: action.list === null || action.list === undefined ? 0 : action.list._page.totalItems,
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
      queryParameters,
    });
  }

  @Action(HomeAction.SearchFail)
  searchDipFail(ctx: StateContext<HomeStateModel>): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @Action(HomeAction.SearchDetail)
  detail(ctx: StateContext<HomeStateModel>, action: HomeAction.SearchDetail): Observable<ArchiveMetadata> {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
    });

    return this.apiService.getById<ArchiveMetadata>(this.getUrlMetadata(ctx), action.resId)
      .pipe(
        tap(model => ctx.dispatch(new HomeAction.SearchDetailSuccess(model))),
        catchError(error => {
          ctx.dispatch(new HomeAction.SearchDetailFail());
          throw error;
        }),
      );
  }

  @Action(HomeAction.SearchDetailSuccess)
  detailSuccess(ctx: StateContext<HomeStateModel>, action: HomeAction.SearchDetailSuccess): void {
    ctx.patchState({
      current: action.model,
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @Action(HomeAction.GetRss)
  getRss(ctx: StateContext<HomeStateModel>, action: HomeAction.GetRss): Observable<Rss> {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
    });

    let headers = new HttpHeaders();
    headers = headers.set("Accept", ["application/xml"]);
    headers = headers.set("Content-Type", "text/xml");

    return this.httpClient.get("/api/rss", {
      headers,
      observe: "response",
      responseType: "text",
    }).pipe(
      map((r: HttpResponse<string>) => {
        const parser = new DOMParser();
        const xml = parser.parseFromString(r.body, "text/xml");
        const obj = this.ngxXml2jsonService.xmlToJson(xml) as { rss: Rss };
        ctx.dispatch(new HomeAction.GetRssSuccess(obj.rss));
        return obj.rss;
      }),
    );
  }

  @Action(HomeAction.GetRssSuccess)
  getRssSuccess(ctx: StateContext<HomeStateModel>, action: HomeAction.GetRssSuccess): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
      rss: action.rss,
    });
  }

  @Action(HomeAction.GetRssFail)
  getRssFail(ctx: StateContext<HomeStateModel>, action: HomeAction.GetRssFail): void {
    this.notificationService.showError(TRANSLATE("welcome.rss.notification.error"), true);
  }

  @Action(HomeAction.Download)
  download(ctx: StateContext<HomeStateModel>, action: HomeAction.Download): any {
    ctx.patchState({
      isLoadingCounterDownloadPreparation: ctx.getState().isLoadingCounterDownloadPreparation + 1,
    });

    let pooling: boolean = true;
    let firstTimeError: boolean = false;

    this.notificationService.showInformation(TRANSLATE("home.archive.browsing.download.notification.inPreparation"), true);

    return interval(500)
      .pipe(
        takeWhile(() => pooling === true),
        startWith(0),
        mergeMap(() => this.getDownloadStatus(ctx, action.dataFile.resId)),
        tap((res: string) => {
          if (res === StringUtil.stringEmpty) {
            this.prepareDownload(ctx, action.dataFile.resId);
          }
          if (res === Order.StatusEnum.READY) {
            this.downloadFile(ctx, action.dataFile.resId);
            pooling = false;
          }
          if (res === Order.StatusEnum.INERROR) {
            pooling = false;
            if (firstTimeError === false) {
              ctx.dispatch(new DownloadFail());
              firstTimeError = true;
            }
          }
        }));

  }

  @Action(HomeAction.DownloadFail)
  downloadFail(ctx: StateContext<HomeStateModel>, action: HomeAction.DownloadFail): void {
    ctx.patchState({
      isLoadingCounterDownloadPreparation: ctx.getState().isLoadingCounterDownloadPreparation - 1,
    });
    this.notificationService.showError(TRANSLATE("home.archive.browsing.download.notification.error"), true);
  }

  saveFile(blobContent: Blob, fileName: string): void {
    const blob = new Blob([blobContent], {type: "application/octet-stream"});
    saveAs(blob, fileName);
  }

  getDownloadStatus(ctx: StateContext<HomeStateModel>, resId: string): Observable<string> {
    let headers = new HttpHeaders();
    headers = headers.set("Content-Type", "text/plain;charset=UTF-8");

    return this.httpClient.get(this.getUrlMetadata(ctx) + urlSeparator + resId
      + urlSeparator + ApiActionEnum.DOWNLOAD_STATUS, {
      headers,
      responseType: "text",
    }).pipe(
      tap((res: string) => res),
      catchError((e/*: ErrorDto*/) => {
        if (e.status === HttpStatus.INTERNAL_SERVER_ERROR) {
          return of(Order.StatusEnum.INERROR);
        }
        return of(StringUtil.stringEmpty);
      }),
    );
  }

  prepareDownload(ctx: StateContext<HomeStateModel>, resId: string): void {
    let headers = new HttpHeaders();
    headers = headers.set("Content-Type", "text/plain;charset=UTF-8");
    this.httpClient.post(this.getUrlMetadata(ctx) + urlSeparator + resId
      + urlSeparator + ApiActionEnum.PREPARE_DOWNLOAD, {headers, responseType: "text"})
      .subscribe();
  }

  downloadFile(ctx: StateContext<HomeStateModel>, resId: string): void {
    ctx.patchState({
      isLoadingCounterDownloadPreparation: ctx.getState().isLoadingCounterDownloadPreparation - 1,
    });

    let headers = new HttpHeaders();
    headers = headers.set("Content-Disposition", "attachment; filename=" + resId);

    this.httpClient.get(`${this.getUrlMetadata(ctx)}/${resId}/${ApiActionEnum.DL}`, {
      headers,
      responseType: "blob",
    }).subscribe((res: Blob) => {
      this.saveFile(res, resId);
    });
  }

}

