import {
  HomeOrganizationalUnitArchiveState,
  HomeOrganizationalUnitArchiveStateModel,
} from "@app/features/home/stores/organizational-unit/archive/home-organizational-unit-archive.state";
import {
  HomeOrgUnitAction,
  homeOrgUnitActionNameSpace,
} from "@app/features/home/stores/organizational-unit/home-organizational-unit.action";
import {AccessOrganizationalUnit} from "@app/generated-api";
import {AccessResourceApiEnum} from "@app/shared/enums/api.enum";
import {LocalStateEnum} from "@app/shared/enums/local-state.enum";
import {RoutesEnum} from "@app/shared/enums/routes.enum";
import {Navigate} from "@ngxs/router-plugin";
import {
  Action,
  Actions,
  State,
  StateContext,
  Store,
} from "@ngxs/store";
import {Observable} from "rxjs";
import {
  catchError,
  tap,
} from "rxjs/operators";
import {
  ApiService,
  CollectionTyped,
  defaultResourceStateInitValue,
  isNullOrUndefined,
  MapUtil,
  NotificationService,
  QueryParameters,
  ResourceState,
  ResourceStateModel,
  StoreUtil,
  TRANSLATE,
} from "solidify-frontend";
import {environment} from "../../../../../environments/environment";

export interface HomeOrganizationalUnitStateModel extends ResourceStateModel<AccessOrganizationalUnit> {
  search: string;
  home_organizationalUnit_archive: HomeOrganizationalUnitArchiveStateModel;
}

@State<HomeOrganizationalUnitStateModel>({
  name: LocalStateEnum.home_organizationalUnit,
  defaults: {
    ...defaultResourceStateInitValue,
    home_organizationalUnit_archive: {isLoadingCounter: 0} as HomeOrganizationalUnitArchiveStateModel,
    queryParameters: new QueryParameters(environment.defaultPageSize),
    search: null,
  },
  children: [
    HomeOrganizationalUnitArchiveState,
  ],
})
export class HomeOrganizationalUnitState extends ResourceState<AccessOrganizationalUnit> {

  private readonly DEFAULT_SEARCH: string = "";

  constructor(protected apiService: ApiService,
              protected store: Store,
              protected notificationService: NotificationService,
              protected actions$: Actions) {
    super(apiService, store, notificationService, actions$, {
      nameSpace: homeOrgUnitActionNameSpace,
      routeRedirectUrlAfterSuccessAction: RoutesEnum.homePage,
      notificationResourceCreateTextToTranslate: TRANSLATE("home.orgunit.notification.resource.create"),
      notificationResourceDeleteTextToTranslate: TRANSLATE("home.orgunit.notification.resource.delete"),
      notificationResourceUpdateTextToTranslate: TRANSLATE("home.orgunit.notification.resource.update"),
    });
  }

  protected get _urlResource(): string {
    return AccessResourceApiEnum.organizationalUnits;
  }

  @Action(HomeOrgUnitAction.Search)
  search(ctx: StateContext<HomeOrganizationalUnitStateModel>, action: HomeOrgUnitAction.Search): void {
    const searchString = this.getSearchStringToApply(ctx, action);
    const queryParameters = StoreUtil.getQueryParametersToApply(action.queryParameters, ctx);

    ctx.patchState({
      search: searchString,
      queryParameters: queryParameters,
    });

    const pathFilter = "name";
    MapUtil.addOrUpdate(queryParameters.search.searchItems, pathFilter, searchString);

    ctx.dispatch(new HomeOrgUnitAction.GetAll(queryParameters));
  }

  private getSearchStringToApply(ctx: StateContext<HomeOrganizationalUnitStateModel>, action: HomeOrgUnitAction.Search): string {
    if (isNullOrUndefined(action.search)) {
      return this.getDefaultSearchStringIfNotDefined(ctx);
    }
    return action.search;
  }

  private getDefaultSearchStringIfNotDefined(ctx: StateContext<HomeOrganizationalUnitStateModel>): string {
    if (ctx.getState().search === null || ctx.getState().search === undefined) {
      return this.DEFAULT_SEARCH;
    }
    return ctx.getState().search;
  }

  @Action(HomeOrgUnitAction.GetDefaultMaximumOrgUnit)
  getDefaultMaximumOrgUnit(ctx: StateContext<ResourceStateModel<HomeOrganizationalUnitStateModel>>, action: HomeOrgUnitAction.GetDefaultMaximumOrgUnit): Observable<CollectionTyped<AccessOrganizationalUnit>> {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
    });

    return this.apiService.get<AccessOrganizationalUnit>(this._urlResource, new QueryParameters(environment.defaultHomeOrgUnitSize))
      .pipe(
        StoreUtil.cancelUncompleted(ctx, this.actions$, [this._nameSpace.GetAll, Navigate]),
        tap((collection: CollectionTyped<AccessOrganizationalUnit>) => {
          ctx.dispatch(new HomeOrgUnitAction.GetDefaultMaximumSuccess(collection));
        }),
        catchError(error => {
          ctx.dispatch(new HomeOrgUnitAction.GetDefaultMaximumFail());
          throw error;
        }),
      );
  }

  @Action(HomeOrgUnitAction.GetDefaultMaximumSuccess)
  getDefaultMaximumSuccess(ctx: StateContext<ResourceStateModel<AccessOrganizationalUnit>>, action: HomeOrgUnitAction.GetDefaultMaximumSuccess): void {
    const queryParameters = StoreUtil.updateQueryParameters(ctx, action.list);
    ctx.patchState({
      list: action.list._data,
      total: action.list._page.totalItems,
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
      queryParameters,
    });
  }

  @Action(HomeOrgUnitAction.GetDefaultMaximumFail)
  getDefaultMaximumFail(ctx: StateContext<ResourceStateModel<AccessOrganizationalUnit>>, action: HomeOrgUnitAction.GetDefaultMaximumFail): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

}
