import {
  ChangeDetectorRef,
  Input,
  OnInit,
  Output,
} from "@angular/core";
import {
  AbstractControl,
  FormGroup,
  ValidationErrors,
  Validators,
} from "@angular/forms";
import {LocalModelAttributeEnum} from "@app/shared/enums/model-attribute.enum";
import {CanDeactivatePresentational} from "@shared/components/presentationals/shared-can-deactivate/shared-can-deactivate.presentational";
import {
  BehaviorSubject,
  Observable,
} from "rxjs";
import {
  FormValidationHelper,
  isNullOrUndefined,
  ModelFormControlEvent,
  ObservableUtil,
} from "solidify-frontend";

export abstract class SharedAbstractFormPresentational<T> extends CanDeactivatePresentational implements OnInit {
  form: FormGroup;

  @Input()
  model: T;

  @Input()
  readonly: boolean;

  isValidWhenDisable: boolean;

  private readonly _submitBS: BehaviorSubject<ModelFormControlEvent<T> | undefined> = new BehaviorSubject<ModelFormControlEvent<T> | undefined>(undefined);

  @Output("submitChange")
  readonly submitObs: Observable<ModelFormControlEvent<T> | undefined> = ObservableUtil.asObservable(this._submitBS);

  get formValidationHelper(): typeof FormValidationHelper {
    return FormValidationHelper;
  }

  required: ValidationErrors = Validators.required;

  protected constructor(protected readonly _changeDetectorRef: ChangeDetectorRef) {
    super();
  }

  ngOnInit(): void {
    super.ngOnInit();
    if (this.model) {
      this.bindFormTo(this.model);
    } else {
      this.initNewForm();
    }
    this.disableIfReadonly();
  }

  private disableIfReadonly(): void {
    if (this.readonly) {
      this.form.disable();
    }
  }

  protected abstract initNewForm(): void;

  protected abstract bindFormTo(model: T): void;

  public onSubmit(): void {
    const model = this.treatmentBeforeSubmit(this.form.value as T);
    this.addResId(model);
    this._submitBS.next({
      model: model,
      formControl: this.form,
      changeDetectorRef: this._changeDetectorRef,
    });
  }

  protected abstract treatmentBeforeSubmit(model: T): T;

  private addResId(model: T): void {
    const isUpdate = this.model !== null && this.model !== undefined;
    if (isUpdate) {
      model[LocalModelAttributeEnum.resId] = this.model[LocalModelAttributeEnum.resId];
    }
  }

  canDeactivate(): boolean {
    return !this.form.dirty;
  }

  getFormControl(key: string): AbstractControl {
    return FormValidationHelper.getFormControl(this.form, key);
  }

  isRequired(key: string): boolean {
    const errors = this.getFormControl(key).errors;
    return isNullOrUndefined(errors) ? false : errors.required;
  }
}
