import {
  ChangeDetectionStrategy,
  Component,
  Input,
  Output,
  ViewChild,
  ViewEncapsulation,
} from "@angular/core";
import {SharedAbstractPresentational} from "@app/shared/components/presentationals/shared-abstract/shared-abstract.presentational";
import {SharedPaginatorPresentational} from "@app/shared/components/presentationals/shared-paginator/shared-paginator.presentational";
import {FieldTypeEnum} from "@app/shared/enums/field-type.enum";
import {DataTableColumns} from "@app/shared/models/data-table-columns.model";
import {LazyLoadEvent} from "primeng/api";
import {
  BehaviorSubject,
  Observable,
} from "rxjs";
import {
  DateUtil,
  isEmptyString,
  isNullOrUndefined,
  isTrue,
  MapUtil,
  ObjectUtil,
  ObservableUtil,
  OrderEnum,
  Paging,
  QueryParameters,
  QueryParametersUtil,
  Sort,
  StringUtil,
} from "solidify-frontend";
import {environment} from "../../../../../environments/environment";

@Component({
  selector: "dlcm-shared-data-table",
  templateUrl: "./shared-data-table.presentational.html",
  styleUrls: ["./shared-data-table.presentational.scss"],
  encapsulation: ViewEncapsulation.None,
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class SharedDataTablePresentational<T> extends SharedAbstractPresentational {
  private _datas: T[];
  defaultSortColumn: DataTableColumns;

  @Input()
  set datas(value: T[]) {
    this._datas = value;
    this.computeIsDatasPresent();
  }

  get datas(): T[] {
    return this._datas;
  }

  @Input()
  skipInitialQuery: boolean = false;

  private _queryParameters: QueryParameters;

  @Input()
  set queryParameters(value: QueryParameters) {
    this._queryParameters = value;
    if (isNullOrUndefined(value.paging)) {
      this._queryParameters = ObjectUtil.clone(this._queryParameters);
      this._queryParameters.paging = new Paging(environment.defaultPageSize);
    }
    if (isNullOrUndefined(value.paging.pageSize)) {
      this._queryParameters = ObjectUtil.clone(this._queryParameters);
      this._queryParameters.paging = ObjectUtil.clone(this._queryParameters.paging);
      this._queryParameters.paging.pageSize = environment.defaultPageSize;
    }
  }

  get queryParameters(): QueryParameters {
    return this._queryParameters;
  }

  private _columns: DataTableColumns[];

  @Input()
  set columns(value: DataTableColumns[]) {
    this._columns = value;
    this.defaultSortColumn = this._columns.find(c => isTrue(c.isSortable) && c.order !== 0);
  }

  get columns(): DataTableColumns[] {
    return this._columns;
  }

  private readonly _selectBS: BehaviorSubject<T | undefined> = new BehaviorSubject<T | undefined>(undefined);
  private readonly _queryParametersBS: BehaviorSubject<QueryParameters | undefined> = new BehaviorSubject<QueryParameters | undefined>(undefined);

  @Output("selectChange")
  readonly selectObs: Observable<T | undefined> = ObservableUtil.asObservable(this._selectBS);

  @Output("queryParametersChange")
  readonly queryParametersObs: Observable<QueryParameters | undefined> = ObservableUtil.asObservable(this._queryParametersBS);

  @ViewChild("paginator", {static: false})
  paginator: SharedPaginatorPresentational;

  isDatasPresent: boolean;

  fieldTypeString: FieldTypeEnum = FieldTypeEnum.string;

  private readonly _SEPARATOR: string = ".";

  constructor() {
    super();
  }

  private computeIsDatasPresent(): void {
    let isDatasPresent = false;
    if (!isNullOrUndefined(this.datas) && this.datas.length > 0) {
      isDatasPresent = true;
    }
    this.isDatasPresent = isDatasPresent;
  }

  showDetail(row: T): void {
    this._selectBS.next(row);
  }

  getCellData(row: T, column: DataTableColumns): any {
    const data = this.getData(row, column);
    return this.applyTypeToData(data, column.type);
  }

  getData(row: T, column: DataTableColumns): any {
    const field = column.field;
    if (field.includes(this._SEPARATOR)) {
      return ObjectUtil.getNestedValue(row, column.field);
    }
    return row[field];
  }

  private applyTypeToData(data: any, type: FieldTypeEnum): any {
    if (isNullOrUndefined(data)) {
      return StringUtil.stringEmpty;
    }
    if (type === FieldTypeEnum.date) {
      return DateUtil.convertDateToDateString(data);
    }
    if (type === FieldTypeEnum.datetime) {
      return DateUtil.convertDateToDateTimeString(data);
    }
    return data;
  }

  backendSort($event: LazyLoadEvent): void {
    if (isTrue(this.skipInitialQuery)) {
      this.skipInitialQuery = false;
      return;
    }
    const queryParameters = ObjectUtil.clone(this.queryParameters);
    if (!isNullOrUndefined($event.sortField)) {
      const sort = this.adaptLazyLoadEventToSortModel($event);
      queryParameters.sort = sort;
    }
    this._queryParametersBS.next(queryParameters);
  }

  private adaptLazyLoadEventToSortModel($event: LazyLoadEvent): Sort {
    return {
      field: $event.sortField,
      order: $event.sortOrder === -1 ? OrderEnum.descending : OrderEnum.ascending,
    } as Sort;
  }

  pageChange($event: Paging): void {
    const queryParameters = ObjectUtil.clone(this.queryParameters);
    queryParameters.paging = $event;
    this._queryParametersBS.next(queryParameters);
  }

  onValueChange(col: DataTableColumns, value: string): void {
    let queryParameters = ObjectUtil.clone(this.queryParameters);
    if (isEmptyString(value)) {
      queryParameters.search.searchItems.delete(col.field);
    } else {
      MapUtil.addOrUpdate(queryParameters.search.searchItems, col.field, value);
    }
    queryParameters = QueryParametersUtil.resetToFirstPage(queryParameters);
    this._queryParametersBS.next(queryParameters);
  }

  getValue(field: string): string {
    const value = this.queryParameters.search.searchItems.get(field);
    if (value === null || value === undefined || value === "") {
      return null;
    }
    return value;
  }
}
