import {PersonExtended} from "@admin/models/person-extended.model";
import {
  Institution,
  Person,
} from "@app/generated-api";
import {AdminResourceApiEnum} from "@app/shared/enums/api.enum";
import {LocalStateEnum} from "@app/shared/enums/local-state.enum";
import {
  SharedPersonAction,
  sharedPersonActionNameSpace,
} from "@app/shared/stores/person/shared-person.action";
import {DepositPersonAction} from "@deposit/stores/people/deposit-person.action";
import {
  Action,
  Actions,
  ofActionCompleted,
  Selector,
  State,
  StateContext,
  Store,
} from "@ngxs/store";
import {ApiActionEnum} from "@shared/enums/api-action.enum";
import {ApiResourceNameEnum} from "@shared/enums/api-resource-name.enum";
import {
  Observable,
  pipe,
} from "rxjs";
import {
  catchError,
  take,
  tap,
} from "rxjs/operators";
import {
  ApiService,
  CollectionTyped,
  defaultResourceStateInitValue,
  isNullOrUndefined,
  isTrue,
  isWhiteString,
  NotificationService,
  ObjectUtil,
  OverrideDefaultAction,
  QueryParameters,
  ResourceState,
  ResourceStateModel,
  StoreUtil,
  TRANSLATE,
  urlSeparator,
} from "solidify-frontend";
import {ActionSubActionCompletionsWrapper} from "solidify-frontend/lib/models/stores/base.action";
import {environment} from "../../../../environments/environment";

export interface SharedPersonStateModel extends ResourceStateModel<PersonExtended> {
  listPersonMatching: PersonExtended[];
}

@State<SharedPersonStateModel>({
  name: LocalStateEnum.shared_person,
  defaults: {
    ...defaultResourceStateInitValue,
    listPersonMatching: [],
    queryParameters: new QueryParameters(environment.defaultEnumValuePageSizeOption),
  },
})
export class SharedPersonState extends ResourceState<PersonExtended> {
  constructor(protected apiService: ApiService,
              protected store: Store,
              protected notificationService: NotificationService,
              protected actions$: Actions) {
    super(apiService, store, notificationService, actions$, {
      nameSpace: sharedPersonActionNameSpace,
      routeRedirectUrlAfterSuccessAction: undefined,
      notificationResourceCreateTextToTranslate: TRANSLATE("shared.person.notification.resource.create"),
      notificationResourceDeleteTextToTranslate: TRANSLATE("shared.person.notification.resource.delete"),
      notificationResourceUpdateTextToTranslate: TRANSLATE("shared.person.notification.resource.update"),
    });
  }

  protected get _urlResource(): string {
    return AdminResourceApiEnum.people;
  }

  @Selector()
  static list<T>(state: ResourceStateModel<T>): T[] {
    return state.list;
  }

  @Selector()
  static total<T>(state: ResourceStateModel<T>): number {
    return state.total;
  }

  @Selector()
  static current<T>(state: ResourceStateModel<T>): T {
    return state.current;
  }

  @Selector()
  static isLoading<T>(state: ResourceStateModel<T>): boolean {
    return StoreUtil.isLoadingState(state);
  }

  @Action(SharedPersonAction.Search)
  search(ctx: StateContext<SharedPersonStateModel>, action: SharedPersonAction.Search): Observable<boolean> {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
    });
    return StoreUtil.dispatchSequentialActionAndWaitForSubActionsCompletion(ctx, [
      {
        action: new SharedPersonAction.SearchPerson(action.person),
        subActionCompletions: [
          this.actions$.pipe(ofActionCompleted(SharedPersonAction.SearchPersonSuccess)),
          this.actions$.pipe(ofActionCompleted(SharedPersonAction.SearchPersonFail)),
        ]
      },
      {
        action: new SharedPersonAction.SearchInstitutions(),
        subActionCompletions: [
          this.actions$.pipe(ofActionCompleted(SharedPersonAction.SearchInstitutionsSuccess)),
          this.actions$.pipe(ofActionCompleted(SharedPersonAction.SearchInstitutionsFail)),
        ]
      }
    ]).pipe(
      tap(success => {
        if (success) {
          ctx.dispatch(new SharedPersonAction.SearchSuccess(action));
        } else {
          ctx.dispatch(new SharedPersonAction.SearchFail(action));
        }
      })
    );
  }

  @Action(SharedPersonAction.SearchSuccess)
  SearchSuccess(ctx: StateContext<SharedPersonStateModel>, action: SharedPersonAction.SearchSuccess): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @Action(SharedPersonAction.SearchFail)
  searchFail(ctx: StateContext<SharedPersonStateModel>, action: SharedPersonAction.SearchFail): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @Action(SharedPersonAction.SearchPerson)
  searchPerson(ctx: StateContext<SharedPersonStateModel>, action: SharedPersonAction.SearchPerson): Observable<CollectionTyped<PersonExtended>> {
    const map = new Map<string, string>();
    map.set("firstName", action.person.firstName);
    map.set("lastName", action.person.lastName);
    const queryParameters = new QueryParameters();
    queryParameters.search = {
      searchItems: map,
    };
    return this.apiService.get<PersonExtended>(this._urlResource, queryParameters)
      .pipe(
        tap(collection => {
          ctx.dispatch(new SharedPersonAction.SearchPersonSuccess(action, collection._data));
        }),
        catchError(error => {
          ctx.dispatch(new SharedPersonAction.SearchPersonFail(action));
          throw error;
        }),
      );
  }

  @Action(SharedPersonAction.SearchPersonSuccess)
  searchPersonSuccess(ctx: StateContext<SharedPersonStateModel>, action: SharedPersonAction.SearchPersonSuccess): void {
    ctx.patchState({
      listPersonMatching: action.list,
    });
  }

  @Action(SharedPersonAction.SearchInstitutions)
  searchInstitutions(ctx: StateContext<SharedPersonStateModel>, action: SharedPersonAction.SearchInstitutions): void {
    const listActionSubActionCompletionsWrapper: ActionSubActionCompletionsWrapper[] = [];

    ctx.getState().listPersonMatching.forEach(p => {
      listActionSubActionCompletionsWrapper.push({
        action: new SharedPersonAction.SearchPersonInstitution(p),
        subActionCompletions: [
          this.actions$.pipe(ofActionCompleted(SharedPersonAction.SearchPersonInstitutionSuccess)),
          this.actions$.pipe(ofActionCompleted(SharedPersonAction.SearchPersonInstitutionFail)),
        ]
      } as ActionSubActionCompletionsWrapper);
    });

    StoreUtil.dispatchParallelActionAndWaitForSubActionsCompletion(ctx, listActionSubActionCompletionsWrapper)
      .pipe(
        tap(success => {
          if (success) {
            ctx.dispatch(new SharedPersonAction.SearchInstitutionsSuccess(action));
          } else {
            ctx.dispatch(new SharedPersonAction.SearchInstitutionsFail(action));
          }
        })
      ).subscribe();
  }

  @Action(SharedPersonAction.SearchPersonInstitution)
  searchPersonInstitutions(ctx: StateContext<SharedPersonStateModel>, action: SharedPersonAction.SearchPersonInstitution): Observable<CollectionTyped<Institution>> {
    return this.apiService.get<Institution>(this._urlResource + urlSeparator + action.person.resId + urlSeparator + ApiResourceNameEnum.INSTITUTION, null)
      .pipe(
        tap(collectionInstitutions => {
          ctx.dispatch(new SharedPersonAction.SearchPersonInstitutionSuccess(action, action.person, collectionInstitutions._data));
        }),
        catchError(error => {
          ctx.dispatch(new SharedPersonAction.SearchPersonInstitutionFail(action));
          throw error;
        }),
      );
  }

  @Action(SharedPersonAction.SearchPersonInstitutionSuccess)
  searchPersonInstitutionsSuccess(ctx: StateContext<SharedPersonStateModel>, action: SharedPersonAction.SearchPersonInstitutionSuccess): void {
    const listPersons: PersonExtended[] = [...ctx.getState().listPersonMatching];
    const personIndex = listPersons.findIndex(p => p.resId === action.person.resId);
    listPersons[personIndex] = ObjectUtil.clone(listPersons[personIndex]);
    listPersons[personIndex].institutions = action.institutions;
    ctx.patchState({
      listPersonMatching: listPersons,
    });
  }

  @OverrideDefaultAction()
  @Action(SharedPersonAction.GetAll)
  getAll(ctx: StateContext<ResourceStateModel<PersonExtended>>, action: SharedPersonAction.GetAll): Observable<CollectionTyped<PersonExtended>> {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
      queryParameters: StoreUtil.getQueryParametersToApply(action.queryParameters, ctx),
    });
    let url = this._urlResource;
    const searchItems = ctx.getState().queryParameters.search.searchItems;
    const fullName = searchItems.get("fullName");
    if (!isNullOrUndefined(fullName) && !isWhiteString(fullName)) {
      searchItems.set("search", `i-firstName~${fullName},i-lastName~${fullName},i-orcid~${fullName}`);
      searchItems.set("match", "any");
      searchItems.delete("fullName");
      url = this._urlResource + urlSeparator + ApiActionEnum.SEARCH;
    }
    return this.apiService.get<PersonExtended>(url, ctx.getState().queryParameters)
      .pipe(
        StoreUtil.cancelUncompleted(ctx, this.actions$, [SharedPersonAction.GetAll]),
        tap((collection: CollectionTyped<PersonExtended>) => {
          ctx.dispatch(new SharedPersonAction.GetAllSuccess(action, collection));
        }),
        catchError(error => {
          ctx.dispatch(new SharedPersonAction.GetAllFail(action));
          throw error;
        }),
      );
  }
}
