import {
  ApplicationRole,
  OrganizationalUnit,
  Person,
  Role,
} from "@app/generated-api";
import {ApplicationRoleEnum} from "@app/shared/enums/application-role.enum";
import {LanguagesEnum} from "@app/shared/enums/languages.enum";
import {RoutesEnum} from "@app/shared/enums/routes.enum";
import {SessionStorageEnum} from "@app/shared/enums/session-storage.enum";
import {ThemeEnum} from "@app/shared/enums/theme.enum";
import {JwtTokenHelper} from "@app/shared/helpers/jwt-token.helper";
import {UserPreferencesUtil} from "@app/shared/utils/user-preferences.util";
import {AppAction} from "@app/stores/app.action";
import {AppAuthorizedOrganizationalUnitAction} from "@app/stores/authorized-organizational-unit/app-authorized-organizational-unit.action";
import {
  AppAuthorizedOrganizationalUnitState,
  AppAuthorizedOrganizationalUnitStateModel,
} from "@app/stores/authorized-organizational-unit/app-authorized-organizational-unit.state";
import {
  AppOrganizationalUnitPersonRoleState,
  AppOrganizationalUnitPersonRoleStateModel,
} from "@app/stores/organizational-unit-person-role/app-organizational-unit-person-role.state";
import {
  AppPersonState,
  AppPersonStateModel,
} from "@app/stores/person/app-person.state";
import {AppUserAction} from "@app/stores/user/app-user.action";
import {
  AppUserState,
  AppUserStateModel,
} from "@app/stores/user/app-user.state";
import {TranslateService} from "@ngx-translate/core";
import {Navigate} from "@ngxs/router-plugin";
import {
  Action,
  Actions,
  ofActionCompleted,
  Selector,
  State,
  StateContext,
  Store,
} from "@ngxs/store";
import {ApiActionEnum} from "@shared/enums/api-action.enum";
import {AdminResourceApiEnum} from "@shared/enums/api.enum";
import {LocalStateEnum} from "@shared/enums/local-state.enum";
import {UserExtended} from "@shared/models/business/user-extended.model";
import {Token} from "@shared/models/token.model";
import * as moment from "moment";
import {Observable} from "rxjs";
import {tap} from "rxjs/operators";
import {
  ApiService,
  isNullOrUndefined,
  isTrue,
  OAuth2Service,
  StoreUtil,
  urlSeparator,
} from "solidify-frontend";
import {environment} from "../../environments/environment";

export interface AppStateModel {
  isApplicationInitialized: boolean;
  isLoggedIn: boolean;
  ignorePreventLeavePopup: boolean;
  userRoles: ApplicationRoleEnum[];
  appLanguage: LanguagesEnum;
  theme: ThemeEnum;
  preventExit: boolean;
  preventExitReasonToTranslate: string | undefined;
  application_user: AppUserStateModel | undefined;
  application_person: AppPersonStateModel | undefined;
  application_authorizedOrganizationalUnit: AppAuthorizedOrganizationalUnitStateModel | undefined;
  application_organizationalUnit_personRole: AppOrganizationalUnitPersonRoleStateModel | undefined;
  token: Token | undefined;
}

@State<AppStateModel>({
  name: LocalStateEnum.application,
  defaults: {
    isApplicationInitialized: false,
    isLoggedIn: false,
    ignorePreventLeavePopup: false,
    userRoles: [ApplicationRoleEnum.guest],
    appLanguage: undefined,
    theme: environment.theme,
    preventExit: false,
    preventExitReasonToTranslate: undefined,
    application_user: undefined,
    application_person: undefined,
    application_authorizedOrganizationalUnit: undefined,
    application_organizationalUnit_personRole: undefined,
    token: undefined,
  },
  children: [
    AppUserState,
    AppPersonState,
    AppAuthorizedOrganizationalUnitState,
    AppOrganizationalUnitPersonRoleState,
  ],
})
export class AppState {
  constructor(private store: Store,
              private translate: TranslateService,
              private oauthService: OAuth2Service,
              private actions$: Actions,
              private apiService: ApiService
              /*private _adapter: DateAdapter<any>*/) {
  }

  @Selector()
  static listAuthorizedOrganizationalUnit(state: AppStateModel): OrganizationalUnit[] {
    return state.application_authorizedOrganizationalUnit.list;
  }

  @Selector()
  static listAuthorizedOrganizationalUnitId(state: AppStateModel): string[] {
    const list = this.listAuthorizedOrganizationalUnit(state);
    if (isNullOrUndefined(list)) {
      return [];
    }
    return list.map(o => o.resId);
  }

  @Selector()
  static currentPerson(state: AppStateModel): Person {
    return state.application_person.current;
  }

  @Selector()
  static currentUser(state: AppStateModel): UserExtended {
    return state.application_user.current;
  }

  @Selector()
  static currentUserApplicationRoles(state: AppStateModel): ApplicationRole[] {
    const currentPersonRole = this.currentUser(state);
    if (isNullOrUndefined(currentPersonRole)) {
      return undefined;
    }
    return currentPersonRole.applicationRoles;
  }

  @Selector()
  static currentUserApplicationRolesResId(state: AppStateModel): string[] | undefined {
    const applicationRoles = this.currentUserApplicationRoles(state);
    if (isNullOrUndefined(applicationRoles)) {
      return undefined;
    }
    return applicationRoles.map(a => a.resId);
  }

  @Selector()
  static currentOrgUnitPersonRole(state: AppStateModel): Role[] | undefined {
    if (isNullOrUndefined(state.application_organizationalUnit_personRole.current)) {
      return undefined;
    }
    return state.application_organizationalUnit_personRole.current.roles;
  }

  @Selector()
  static currentOrgUnitPersonRoleResId(state: AppStateModel): string[] | undefined {
    const currentPersonRole = this.currentOrgUnitPersonRole(state);
    if (isNullOrUndefined(currentPersonRole)) {
      return undefined;
    }
    return currentPersonRole.map(r => r.resId);
  }

  @Action(AppAction.InitApplication)
  initApplication(ctx: StateContext<AppStateModel>, action: AppAction.InitApplication): void {
    StoreUtil.dispatchParallelActionAndWaitForSubActionsCompletion(
      ctx,
      [
        {
          action: new AppAction.Login(),
          subActionCompletions: [
            this.actions$.pipe(ofActionCompleted(AppAction.LoginSuccess)),
            this.actions$.pipe(ofActionCompleted(AppAction.LoginFail)),
          ],
        },
        {action: new AppAction.ChangeAppLanguage(UserPreferencesUtil.getPreferredLanguage())},
        {action: new AppAction.DefineFallbackLanguage()},
        {action: new AppAction.SetMomentLocal()},
      ],
    ).subscribe(isAuthenticatedUser => {
      ctx.dispatch(new AppAction.InitApplicationSuccess(action));
    });
  }

  @Action(AppAction.InitApplicationSuccess)
  initApplicationSuccess(ctx: StateContext<AppStateModel>): void {
    ctx.patchState({
      isApplicationInitialized: true,
    });
  }

  @Action(AppAction.DefineFallbackLanguage)
  defineFallbackLanguage(ctx: StateContext<AppStateModel>): void {
    // this language will be used as a fallback when a translation isn't found in the current language
    this.translate.setDefaultLang(environment.defaultLanguage);
  }

  @Action(AppAction.SetMomentLocal)
  setMomentLocal(ctx: StateContext<AppStateModel>): void {
    const navigatorLanguage = navigator.language;
    moment.locale(navigatorLanguage);
    // this._adapter.setLocale("fr-CH"); // TOFIX : DOESN'T WORK HERE, NEED TO BE SET IN APP MODULE
  }

  @Action(AppAction.ChangeAppLanguage)
  changeAppLanguage(ctx: StateContext<AppStateModel>, action: AppAction.ChangeAppLanguage): Observable<any> {
    ctx.patchState({
      appLanguage: action.language,
    });

    sessionStorage.setItem(SessionStorageEnum.language, action.language);
    return this.translate.use(action.language);
  }

  @Action(AppAction.ChangeAppTheme)
  changeAppTheme(ctx: StateContext<AppStateModel>, action: AppAction.ChangeAppTheme): void {
    environment.theme = action.theme;
    ctx.patchState({
      theme: action.theme,
    });
  }

  @Action(AppAction.Login)
  login(ctx: StateContext<AppStateModel>, action: AppAction.Login): Observable<boolean> {
    return this.oauthService.tryLogin()
      .pipe(
        tap(e => {
            if (e) {
              ctx.dispatch(new AppAction.LoginSuccess(action));
            } else {
              ctx.dispatch(new AppAction.LoginFail(action));
            }
          },
        ));
  }

  @Action(AppAction.LoginSuccess)
  loginSuccess(ctx: StateContext<AppStateModel>): void {
    const accessToken = this.oauthService.getAccessToken();
    const tokenDecoded = JwtTokenHelper.decodeToken(accessToken);
    ctx.patchState({
      isLoggedIn: true,
      userRoles: tokenDecoded.authorities,
      token: tokenDecoded,
    });
    StoreUtil.dispatchParallelActionAndWaitForSubActionsCompletion(ctx, [
      {
        action: new AppUserAction.GetCurrentUser(),
        subActionCompletions: [
          this.actions$.pipe(ofActionCompleted(AppUserAction.GetCurrentUserSuccess)),
          this.actions$.pipe(ofActionCompleted(AppUserAction.GetCurrentUserFail)),
        ],
      },
      {
        action: new AppAuthorizedOrganizationalUnitAction.GetAll(),
        subActionCompletions: [
          this.actions$.pipe(ofActionCompleted(AppAuthorizedOrganizationalUnitAction.GetAllSuccess)),
          this.actions$.pipe(ofActionCompleted(AppAuthorizedOrganizationalUnitAction.GetAllFail)),
        ],
      },
    ]).subscribe(success => {

    });
  }

  @Action(AppAction.LoginFail)
  loginFail(ctx: StateContext<AppStateModel>): void {
    ctx.patchState({
      isLoggedIn: false,
    });
  }

  @Action(AppAction.Logout)
  logout(ctx: StateContext<AppStateModel>): void {
    // Revoke tokens servers side
    this.apiService.post(AdminResourceApiEnum.users + urlSeparator + ApiActionEnum.REVOKE_MY_TOKENS).subscribe();
    // Logout client side
    this.oauthService.logOut();
    ctx.patchState({
      ignorePreventLeavePopup: true,
      isLoggedIn: false,
      userRoles: [ApplicationRoleEnum.guest],
      token: undefined,
      application_user: undefined,
    });
    ctx.dispatch(new Navigate([RoutesEnum.homePage])).subscribe(() => {
      ctx.patchState({
        ignorePreventLeavePopup: false,
      });
      if (isTrue(environment.reloadAfterLogout)) {
        window.location.reload();
      }
    });
  }

  @Action(AppAction.CancelPreventExit)
  cancelPreventExit(ctx: StateContext<AppStateModel>): void {
    ctx.patchState({
      preventExit: false,
      preventExitReasonToTranslate: undefined,
    });
  }

  @Action(AppAction.PreventExit)
  preventExit(ctx: StateContext<AppStateModel>, action: AppAction.PreventExit): void {
    ctx.patchState({
      preventExit: true,
      preventExitReasonToTranslate: action.reasonToTranslate,
    });
  }
}
